from __future__ import annotations

import os
import sqlite3
from typing import Any, Mapping, Type

import numpy as np
import pandas as pd
import requests


def revdict(d: dict) -> dict:
    return {v: k for k, v in d.items()}


COLUMN_TYPES = {
    "year_of_birth": np.int16,
    "year_of_death": np.int16,
    "gender_id": np.int8,
    "party_id": np.int8,
    "chamber_id": np.int8,
    "office_type_id": np.int8,
    "sub_office_type_id": np.int8,
    "start_year": np.int16,
    "end_year": np.int16,
    "district_id": np.int16,
}

COLUMN_DEFAULTS = {
    "gender_id": 0,
    "year_of_birth": 0,
    "year_of_death": 0,
    "district_id": 0,
    "party_id": 0,
    "chamber_id": 0,
    "office_type_id": 0,
    "sub_office_type_id": 0,
    "start_year": 0,
    "end_year": 0,
}


PARTY_COLORS = [
    (0, "S", "#E8112d"),
    (1, "M", "#52BDEC"),
    (2, "gov", "#000000"),
    (3, "C", "#009933"),
    (4, "L", "#006AB3"),
    (5, "V", "#DA291C"),
    (6, "MP", "#83CF39"),
    (7, "KD", "#000077"),
    (8, "NYD", "#007700"),
    (9, "SD", "#DDDD00"),
]

PARTY_COLOR_BY_ID = {x[0]: x[2] for x in PARTY_COLORS}
PARTY_COLOR_BY_ABBREV = {x[1]: x[2] for x in PARTY_COLORS}

NAME2IDNAME_MAPPING: Mapping[str, str] = {
    "gender": "gender_id",
    "office_type": "office_type_id",
    "sub_office_type": "sub_office_type_id",
    "person_id": "pid",
}
IDNAME2NAME_MAPPING: Mapping[str, str] = revdict(NAME2IDNAME_MAPPING)


def read_sql_table(table_name: str, con: sqlite3.Connection) -> pd.DataFrame:
    return pd.read_sql(f"select * from {table_name}", con)


def read_sql_tables(
    tables: list[str] | dict, db: sqlite3.Connection
) -> dict[str, pd.DataFrame]:
    return (
        tables
        if isinstance(tables, dict)
        else {table_name: read_sql_table(table_name, db) for table_name in tables}
    )


def load_tables(
    tables: dict[str, str],
    *,
    db: sqlite3.Connection,
    defaults: dict[str, Any] = None,
    types: dict[str, Any] = None,
):
    """Loads tables as pandas dataframes, slims types, fills NaN, sets pandas index"""
    data: dict[str, pd.DataFrame] = read_sql_tables(list(tables.keys()), db)
    slim_table_types(data.values(), defaults=defaults, types=types)
    for table_name, table in data.items():
        if tables.get(table_name):
            table.set_index(tables.get(table_name), drop=True, inplace=True)
    return data


def slim_table_types(
    tables: list[pd.DataFrame] | pd.DataFrame,
    defaults: dict[str, Any] = None,
    types: dict[str, Any] = None,
) -> None:
    """Slims types and sets default value for NaN entries"""

    if isinstance(tables, pd.DataFrame):
        tables = [tables]

    defaults: dict[str, Any] = COLUMN_DEFAULTS if defaults is None else defaults
    types: dict[str, Any] = COLUMN_TYPES if types is None else types

    for table in tables:
        for column_name, value in defaults.items():
            if column_name in table.columns:
                table[column_name].fillna(value, inplace=True)

        for column_name, dt in types.items():
            if column_name in table.columns:
                if table[column_name].dtype != dt:
                    table[column_name] = table[column_name].astype(dt)


def group_to_list_of_records2(
    df: pd.DataFrame, key: str
) -> dict[str | int, list[dict]]:
    """Groups `df` by `key` and aggregates each group to list of row records (dicts)"""
    return {
        q: df.loc[ds].to_dict(orient="records")
        for q, ds in df.groupby(key).groups.items()
    }


def group_to_list_of_records(
    df: pd.DataFrame, key: str, properties: list[str] = None, ctor: Type = None
) -> dict[str | int, list[dict]]:
    """Groups `df` by `key` and aggregates each group to list of row records (dicts)"""
    key_rows: pd.DataFrame = pd.DataFrame(
        data={
            key: df[key],
            "data": (df[properties] if properties else df).to_dict("records"),
        }
    )
    if ctor is not None:
        key_rows["data"] = key_rows["data"].apply(lambda x: ctor(**x))

    return key_rows.groupby(key)["data"].apply(list).to_dict()


def download_url_to_file(url: str, target_name: str, force: bool = False) -> None:
    if os.path.isfile(target_name):
        if not force:
            raise ValueError("File exists, use `force=True` to overwrite")
        os.unlink(target_name)

    ensure_path(target_name)

    with open(target_name, "w", encoding="utf-8") as fp:
        data: str = requests.get(url, allow_redirects=True).content.decode("utf-8")
        fp.write(data)


def probe_filename(filename: list[str], exts: list[str] = None) -> str | None:
    """Probes existence of filename with any of given extensions in folder"""
    for probe_name in set(
        [filename]
        + ([replace_extension(filename, ext) for ext in exts] if exts else [])
    ):
        if os.path.isfile(probe_name):
            return probe_name
    raise FileNotFoundError(filename)


def replace_extension(filename: str, extension: str) -> str:
    if filename.endswith(extension):
        return filename
    base, _ = os.path.splitext(filename)
    return f"{base}{'' if extension.startswith('.') else '.'}{extension}"


def ensure_path(f: str) -> None:
    os.makedirs(os.path.dirname(f), exist_ok=True)



from __future__ import annotations

import sqlite3
from contextlib import nullcontext
from dataclasses import dataclass
from functools import cached_property
from typing import Callable, Literal, Mapping

import pandas as pd
from penelope import utility as pu  # type: ignore

from .utility import load_tables

CODE_TABLENAMES: dict[str, str] = {
    "chamber": "chamber_id",
    "gender": "gender_id",
    "government": "government_id",
    "office_type": "office_type_id",
    "party": "party_id",
    "sub_office_type": "sub_office_type_id",
}


@dataclass
class Codec:
    type: Literal["encode", "decode"]
    from_column: str
    to_column: str
    fx: Callable[[int], str]
    default: str = None


null_frame: pd.DataFrame = pd.DataFrame()


class Codecs:
    def __init__(self):
        self.chamber: pd.DataFrame = null_frame
        self.gender: pd.DataFrame = null_frame
        self.government: pd.DataFrame = null_frame
        self.office_type: pd.DataFrame = null_frame
        self.party: pd.DataFrame = null_frame
        self.sub_office_type: pd.DataFrame = null_frame
        self.extra_codecs: list[Codec] = []
        self.source_filename: str | None = None

    def load(self, source: str | sqlite3.Connection | str) -> Codecs:
        self.source_filename = source if isinstance(source, str) else None
        with (
            sqlite3.connect(database=source)
            if isinstance(source, str)
            else nullcontext(source)
        ) as db:
            tables: dict[str, pd.DataFrame] = load_tables(self.tablenames(), db=db)
            for table_name, table in tables.items():
                setattr(self, table_name, table)
        return self

    def tablenames(self) -> dict[str, str]:
        """Returns a mapping from code table name to id column name"""
        return CODE_TABLENAMES

    @cached_property
    def gender2name(self) -> dict:
        """Returns a mapping from gender ID to gender name"""
        return self.gender["gender"].to_dict()

    @cached_property
    def gender2id(self) -> dict:
        """Returns a mapping from gender name to gender ID"""
        return pu.revdict(self.gender2name)

    @cached_property
    def office_type2name(self) -> dict:
        return self.office_type["office"].to_dict()

    @cached_property
    def office_type2id(self) -> dict:
        return pu.revdict(self.office_type2name)

    @cached_property
    def sub_office_type2name(self) -> dict:
        return self.sub_office_type["description"].to_dict()

    @cached_property
    def sub_office_type2id(self) -> dict:
        return pu.revdict(self.sub_office_type2name)

    @cached_property
    def party_abbrev2name(self) -> dict:
        return self.party["party_abbrev"].to_dict()

    @cached_property
    def party_abbrev2id(self) -> dict:
        return pu.revdict(self.party_abbrev2name)

    @property
    def codecs(self) -> list[Codec]:
        return self.extra_codecs + [
            Codec("decode", "gender_id", "gender", self.gender2name.get),
            Codec("decode", "office_type_id", "office_type", self.office_type2name.get),
            Codec("decode", "party_id", "party_abbrev", self.party_abbrev2name.get),
            Codec(
                "decode",
                "sub_office_type_id",
                "sub_office_type",
                self.sub_office_type2name.get,
            ),
            Codec("encode", "gender", "gender_id", self.gender2id.get),
            Codec("encode", "office_type", "office_type_id", self.office_type2id.get),
            Codec("encode", "party", "party_id", self.party_abbrev2id.get),
            Codec(
                "encode",
                "sub_office_type",
                "sub_office_type_id",
                self.sub_office_type2id.get,
            ),
        ]

    @property
    def decoders(self) -> list[Codec]:
        return [c for c in self.codecs if c.type == "decode"]

    @property
    def encoders(self) -> list[dict]:
        return [c for c in self.codecs if c.type == "encode"]

    def apply_codec(
        self, df: pd.DataFrame, codecs: list[Codec], drop: bool = True
    ) -> pd.DataFrame:
        for codec in codecs:
            if codec.from_column in df.columns:
                if codec.to_column not in df:
                    df[codec.to_column] = df[codec.from_column].apply(codec.fx)
                if codec.default is not None:
                    df[codec.to_column] = df[codec.to_column].fillna(codec.default)
            if drop:
                df.drop(columns=[codec.from_column], inplace=True, errors="ignore")
        return df

    def decode(self, df: pd.DataFrame, drop: bool = True) -> pd.DataFrame:
        return self.apply_codec(df, self.decoders, drop=drop)

    def encode(self, df: pd.DataFrame, drop: bool = True) -> pd.DataFrame:
        return self.apply_codec(df, self.encoders, drop=drop)

    @cached_property
    def property_values_specs(self) -> list[Mapping[str, str | Mapping[str, int]]]:
        return [
            dict(text_name="gender", id_name="gender_id", values=self.gender2id),
            dict(
                text_name="office_type",
                id_name="office_type_id",
                values=self.office_type2id,
            ),
            dict(
                text_name="party_abbrev",
                id_name="party_id",
                values=self.party_abbrev2id,
            ),
            dict(
                text_name="sub_office_type",
                id_name="sub_office_type_id",
                values=self.sub_office_type2id,
            ),
        ]

    @cached_property
    def key_name_translate_id2text(self) -> dict:
        return {
            codec.from_column: codec.to_column
            for codec in self.codecs
            if codec.type == "decode"
        }

    @cached_property
    def key_name_translate_text2id(self) -> dict:
        return pu.revdict(self.key_name_translate_id2text)

    @cached_property
    def key_name_translate_any2any(self) -> dict:
        """Translates key's id/text name to corresponding text (id) name e.g. `gender_id` => `gender`"""
        translation: dict = {}
        translation.update(self.key_name_translate_id2text)
        translation.update(self.key_name_translate_text2id)
        return translation

    def translate_key_names(self, keys: list[str]) -> list[str]:
        """Translates keys' id/text name to corresponding text (id) name e.g. `gender_id` => `gender`"""
        fg = self.key_name_translate_any2any.get
        return [fg(key) for key in keys if fg(key) is not None]


class PersonCodecs(Codecs):
    def __init__(self):
        super().__init__()
        self.persons_of_interest: pd.DataFrame = null_frame

    def tablenames(self) -> dict[str, str]:
        tables: dict[str, str] = dict(CODE_TABLENAMES)
        tables["persons_of_interest"] = "person_id"
        return tables

    def load(self, source: str | sqlite3.Connection | dict) -> Codecs:
        super().load(source)
        if "pid" not in self.persons_of_interest.columns:
            pi: pd.DataFrame = self.persons_of_interest.reset_index()
            pi["pid"] = pi.index
            pi.set_index("person_id", inplace=True)
            self.persons_of_interest = pi
        return self

    @cached_property
    def pid2person_id(self) -> dict:
        return self.person.reset_index().set_index("pid")["person_id"].to_dict()

    @cached_property
    def person_id2pid(self) -> dict:
        return pu.revdict(self.pid2person_id)

    @cached_property
    def pid2person_name(self) -> dict:
        return self.person.reset_index().set_index("pid")["name"].to_dict()

    @cached_property
    def person_name2pid(self) -> dict:
        fg = self.person_id2pid.get
        return {
            f"{name} ({person_id})": fg(person_id)
            for person_id, name in self.person_id2name.items()
        }

    @cached_property
    def property_values_specs(self) -> list[Mapping[str, str | Mapping[str, int]]]:
        return super().property_values_specs + [
            dict(text_name="name", id_name="pid", values=self.person_name2pid),
        ]

    @cached_property
    def person_id2name(self) -> dict[str, str]:
        fg = self.pid2person_id.get
        return {fg(pid): name for pid, name in self.pid2person_name.items()}

    @property
    def person(self) -> pd.DataFrame:
        return self.persons_of_interest

    @property
    def codecs(self) -> list[Codec]:
        return (
            self.extra_codecs
            + super().codecs
            + [
                Codec("decode", "person_id", "name", self.person_id2name.get),
                Codec("decode", "pid", "person_id", self.pid2person_id.get),
                Codec("encode", "person_id", "pid", self.person_id2pid.get),
            ]
        )



from __future__ import annotations

from typing import Any

import pandas as pd
import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore


class SpeakerDropDown:
    def __init__(
        self, another_api: ADummyApi, party_selections: dict, gender_selections: dict
    ) -> None:
        self.api = another_api
        self.speakers = self.api.decoded_persons
        self.party_selections = party_selections
        self.gender_selections = gender_selections
        self.selector = self.draw_multiselect()

    def draw_multiselect(self) -> Any:
        current_speakers = self.get_speakers_for_selected_parties()
        if self.gender_selections is not None:
            current_speakers = self.get_speakers_selected_genders(current_speakers)
        return st.multiselect(
            "Sök enskilda talare",
            current_speakers.index,
            format_func=self.format_speaker,
            default=[],
            key="speaker_dropdown",
        )

    def get_speakers_selected_genders(
        self, current_speakers: pd.DataFrame
    ) -> pd.DataFrame:
        return current_speakers[
            current_speakers["gender_id"].isin(self.gender_selections["gender_id"])
        ]

    def get_speakers_for_selected_parties(self) -> pd.DataFrame:
        if self.party_selections is not None:
            return self.speakers[
                self.speakers["party_id"].isin(self.party_selections["party_id"])
            ]
        return self.speakers

    def format_speaker(self, x: str) -> str:
        if (birth_year := self.speakers.loc[x, "year_of_birth"]) != 0:
            birth_year = f"{birth_year} - "
        else:
            birth_year = ""
        if (death_year := self.speakers.loc[x, "year_of_death"]) == 0:
            death_year = ""
        name = self.speakers.loc[x, "name"]
        party = self.speakers.loc[x, "party_abbrev"]
        return f'{name} ({party}) {birth_year} {death_year}'

    def get_selection(self) -> dict | None:
        if len(self.selector) == 0:
            return None
        return {"who": self.selector}

    def get_selected_speakers_string(self) -> str:
        if len(self.selector) == 0:
            return "Alla talare"
        selected = ', '.join([self.format_speaker(x) for x in self.selector])
        return f"**Valda talare**: {selected}.\n"



import sys

from penelope.utility.paths import (find_data_folder, find_resources_folder,
                                    find_root_folder)

project_name: str = "sample-data"
project_short_name: str = "swedeb"

corpus_folder: str = find_data_folder(
    project_name=project_name, project_short_name=project_short_name
)
root_folder: str = find_root_folder(project_name=project_name)
resources_folder: str = find_resources_folder(
    project_name=project_name, project_short_name=project_short_name
)

data_folder: str = corpus_folder

if root_folder not in sys.path:
    sys.path.insert(0, root_folder)



from __future__ import annotations

from typing import Any, Optional

import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore
from swedeb_demo.components.gender_checkboxes import GenderCheckBoxes
from swedeb_demo.components.party_dropdown import PartyDropDown
from swedeb_demo.components.speaker_dropdown import SpeakerDropDown


# tid, parti, kön, (kammare, valkrets), namn
class MetaDataDisplay:
    def __init__(self, st_dict: Any, api: ADummyApi) -> None:
        self.st_dict = st_dict
        self.another_api = api
        self.min_year = self.another_api.get_years_start()
        self.max_year = self.another_api.get_years_end()
        self.add_filter_components()

    def add_filter_components(self) -> None:
        self.slider = st.slider(
            "Välj intervall:",
            value=[self.min_year, self.max_year],
            min_value=self.min_year,
            max_value=self.max_year,
            key="years_slider",
        )
        self.column_container = st.container()

        with self.column_container:
            self.gender_boxes = GenderCheckBoxes()
            self.party_dropdown = PartyDropDown(self.another_api)
            self.speaker_dropdown = SpeakerDropDown(
                self.another_api,
                self.party_dropdown.get_selection(),
                self.gender_boxes.get_selection(),
            )

    def get_slider(self) -> Any:
        return self.slider

    def get_gender_selection(self) -> Optional[dict]:
        return self.gender_boxes.get_selection()

    def get_party_selection(self) -> dict | None:
        return self.party_dropdown.get_selection()

    def get_speaker_selection(self) -> dict | None:
        return self.speaker_dropdown.get_selection()

    def get_selections(self) -> dict:
        selections = {}
        gs = self.gender_boxes.get_selection()
        ps = self.party_dropdown.get_selection()
        ss = self.speaker_dropdown.get_selection()
        if gs is not None:
            selections.update(gs)
        if ps is not None:
            selections.update(ps)
        if ss is not None:
            selections.update(ss)
        return selections

    def get_selected_years_str(self) -> str:
        selected_years = self.get_slider()
        return f"**Intervall:** {selected_years[0]} – {selected_years[1]}"

    def get_current_settings(self) -> str:
        if self.speaker_dropdown.get_selected_speakers_string() == "Alla talare":
            return " ".join(
                [
                    self.gender_boxes.get_selected_genders_string(),
                    self.party_dropdown.get_selected_parties_string(),
                    self.get_selected_years_str(),
                ]
            )
        else:
            return " ".join(
                [
                    self.speaker_dropdown.get_selected_speakers_string(),
                    self.get_selected_years_str(),
                ]
            )



from dataclasses import dataclass

import pandas as pd
from penelope import corpus as pc  # type: ignore
from penelope.common.keyness.metrics import KeynessMetric  # type: ignore
from penelope.notebook import word_trends as wt  # type: ignore

from . import codecs as md

# These two class are currently identical to the ones in welfare_state_analytics.notebookd...word_trends.py


@dataclass
class SweDebComputeOpts(wt.TrendsComputeOpts):
    source_folder: str = None

    def invalidates_corpus(self, other: "SweDebComputeOpts") -> bool:
        if super().invalidates_corpus(other):
            return True
        if self.source_folder != other.source_folder:
            return True
        return False

    @property
    def clone(self) -> "SweDebComputeOpts":
        obj: SweDebComputeOpts = super(
            SweDebComputeOpts, self
        ).clone  # pylint: disable=super-with-arguments
        obj.source_folder = self.source_folder
        return obj


class SweDebTrendsData(wt.TrendsService):
    def __init__(
        self,
        corpus: pc.VectorizedCorpus,
        person_codecs: md.PersonCodecs,
        n_top: int = 100000,
    ):
        super().__init__(corpus, n_top=n_top)
        self.person_codecs: md.PersonCodecs = person_codecs
        self._compute_opts: SweDebComputeOpts = SweDebComputeOpts(
            normalize=False,
            keyness=KeynessMetric.TF,
            temporal_key="decade",
            top_count=None,
            words=None,
        )

    def _transform_corpus(self, opts: SweDebComputeOpts) -> pc.VectorizedCorpus:
        corpus: pc.VectorizedCorpus = super()._transform_corpus(opts)
        di: pd.DataFrame = self.update_document_index(opts, corpus.document_index)
        corpus.replace_document_index(di)
        return corpus

    def update_document_index(
        self, opts: SweDebComputeOpts, document_index: pd.DataFrame
    ) -> pd.DataFrame:
        """Decodes ID columns (keeps ID) and updates document index with filename, time_period and document_name."""
        if not opts.pivot_keys_id_names:
            return document_index
        di: pd.DataFrame = self.person_codecs.decode(document_index, drop=False)
        pivot_keys_text_names = self.person_codecs.translate_key_names(
            opts.pivot_keys_id_names
        )
        di["document_name"] = di[pivot_keys_text_names].apply(
            lambda x: "_".join(x).lower(), axis=1
        )
        di["filename"] = di.document_name
        di["time_period"] = di[opts.temporal_key]
        return di



from __future__ import annotations

import abc
import json
import os
import re
import sqlite3
import types
import zipfile
from collections import namedtuple
from functools import cached_property
from os.path import join as jj
from typing import Iterable, Literal

import numpy as np
import pandas as pd
from jinja2 import Template
from loguru import logger
from penelope import utility as pu

from . import codecs as md
from .utility import read_sql_table

try:
    import github as gh
except ImportError:
    Github = lambda t: types.SimpleNamespace()


default_template: Template = Template(
    """
<b>Protokoll:</b> {{protocol_name}} sidan {{ page_number }}, {{ chamber }}, {{ date }} <br/>
<b>Källor:</b> {{parlaclarin_links}} {{ wikidata_link }} {{ kb_labb_link }} <br/>
<b>Talare:</b> {{name}}, {{ party_abbrev }}, {{ office_type }}, {{ sub_office_type }}, {{ district }}, {{ gender}}<br/>
<b>Antal tokens:</b> {{ num_tokens }} ({{ num_words }}), uid: {{u_id}}, who: {{who}} <br/>
<h3> {{ speaker_note }} </h3>
<span style="color: blue;line-height:50%;">
{% for n in paragraphs %}
{{n}}
{% endfor %}
</span>
"""
)

GithubUrl = namedtuple("GithubUrl", "name url")

# pylint: disable=unused-argument


class SpeechTextService:
    """Reconstitute text using information stored in the document (speech) index"""

    def __init__(self, document_index: pd.DataFrame):
        self.speech_index: pd.DataFrame = document_index
        self.speech_index["protocol_name"] = (
            self.speech_index["document_name"].str.split("_").str[0]
        )
        self.speech_index.rename(
            columns={"speach_index": "speech_index"}, inplace=True, errors="ignore"
        )

        """Name of speaker note reference was changed from v0.4.3 (speaker_hash => speaker_note_id)"""
        self.id_name = (
            "speaker_note_id"
            if "speaker_note_id" in self.speech_index.columns
            else "speaker_hash"
        )

    @cached_property
    def name2info(self) -> dict[str, dict]:
        """Create a map protcol name to list of dict of relevant properties"""
        si: pd.DataFrame = self.speech_index.set_index("protocol_name", drop=True)[
            ["u_id", "speech_index", self.id_name, "n_utterances"]
        ]
        return (
            si.assign(data=si.to_dict("records"))
            .groupby(si.index)
            .agg(list)["data"]
            .to_dict()
        )

    def speeches(self, *, metadata: dict, utterances: list[dict]) -> list[dict]:
        """Create list of speeches for all speeches in protocol"""
        speech_infos: dict = self.name2info.get(metadata.get("name"))
        speech_lengths: np.ndarray = np.array(
            [s.get("n_utterances", 0) for s in speech_infos]
        )
        speech_starts: np.ndarray = np.append([0], np.cumsum(speech_lengths))
        speeches = [
            self._create_speech(
                metadata=metadata,
                utterances=utterances[speech_starts[i] : speech_starts[i + 1]],
            )
            for i in range(0, len(speech_infos))
        ]
        return speeches

    def nth(self, *, metadata: dict, utterances: list[dict], n: int) -> dict:
        # speech_infos: dict = self.name2info.get(metadata.get("name"))
        # u_idx: int = [u['u_id'] for u in utterances].index(speech_infos['u_id'])
        # self._create_speech(metadata, utterances[u_idx:u_idx+speech_infos['n_utterances']])
        return self.speeches(metadata=metadata, utterances=utterances)[n]

    def _create_speech(self, *, metadata: dict, utterances: list[dict]) -> dict:
        return (
            {}
            if len(list(utterances or [])) == 0
            else dict(
                speaker_note_id=utterances[0][self.id_name],
                who=utterances[0]["who"],
                u_id=utterances[0]["u_id"],
                paragraphs=[p for u in utterances for p in u["paragraphs"]],
                num_tokens=sum(x["num_tokens"] for x in utterances),
                num_words=sum(x["num_words"] for x in utterances),
                page_number=utterances[0]["page_number"] or "?",
                page_number2=utterances[-1]["page_number"] or "?",
                protocol_name=(metadata or {}).get("name", "?"),
                date=(metadata or {}).get("date", "?"),
            )
        )


class Loader(abc.ABC):
    @abc.abstractmethod
    def load(self, protocol_name: str) -> tuple[dict, list[dict]]:
        ...


class ZipLoader(Loader):
    def __init__(self, folder: str):
        self.folder: str = folder

    def load(self, protocol_name: str) -> tuple[dict, list[dict]]:
        """Loads tagged protocol data from archive"""
        sub_folder: str = protocol_name.split("-")[1]
        for filename in [
            jj(self.folder, sub_folder, f"{protocol_name}.zip"),
            jj(self.folder, f"{protocol_name}.zip"),
        ]:
            if not os.path.isfile(filename):
                continue
            with zipfile.ZipFile(filename, "r") as fp:
                json_str: str = fp.read(f"{protocol_name}.json")
                metadata_str: str = fp.read("metadata.json")
            metadata: dict = json.loads(metadata_str)
            utterances: list[dict] = json.loads(json_str)
            return metadata, utterances
        raise FileNotFoundError(protocol_name)


class SpeechTextRepository:
    GITHUB_REPOSITORY_URL: str = (
        "https://github.com/welfare-state-analytics/riksdagen-corpus"
    )
    GITHUB_REPOSITORY_RAW_URL = (
        "https://raw.githubusercontent.com/welfare-state-analytics/riksdagen-corpus"
    )

    def __init__(
        self,
        *,
        source: str | Loader,
        person_codecs: md.PersonCodecs,
        document_index: pd.DataFrame,
        template: Template = None,
        service: SpeechTextService = None,
    ):
        self.template: Template = template or default_template
        self.source: Loader = (
            source if isinstance(source, Loader) else ZipLoader(source)
        )
        self.person_codecs: md.PersonCodecs = person_codecs
        self.document_index: pd.DataFrame = document_index
        self.subst_puncts = re.compile(r'\s([,?.!"%\';:`](?:\s|$))')
        self.release_tags: list[str] = self.get_github_tags()
        self.service: SpeechTextService = service or SpeechTextService(
            self.document_index
        )
        self.document_name2id: dict[str, int] = (
            document_index.reset_index()
            .set_index("document_name")["document_id"]
            .to_dict()
        )

    def load_protocol(self, protocol_name: str) -> tuple[dict, list[dict]]:
        return self.source.load(protocol_name)

    def speeches(self, protocol_name: str) -> Iterable[dict]:
        metadata, utterances = self.source.load(protocol_name)
        return self.service.speeches(utterances=utterances, metadata=metadata)

    def _get_speech_info(self, speech_id: int | str) -> dict:
        """Get speaker-info from document index and person table"""

        speech_id: int = (
            self.document_name2id.get(speech_id)
            if isinstance(speech_id, str)
            else speech_id
        )

        try:
            speech_info: dict = self.document_index.loc[speech_id].to_dict()
        except KeyError as ex:
            raise KeyError(f"Speech {speech_id} not found in index") from ex

        try:
            speaker_name: str = (
                self.person_codecs.person.loc[speech_info["who"]]["name"]
                if speech_info
                else "unknown"
            )
        except KeyError:
            speaker_name: str = speech_info["who"]

        speech_info.update(name=speaker_name)

        speech_info["speaker_note"] = self.speaker_note_id2note.get(
            speech_info.get(self.service.id_name), "(introductory note not found)"
        )

        return speech_info

    @cached_property
    def speaker_note_id2note(self) -> dict:
        try:
            if not self.person_codecs.source_filename:
                return {}
            with sqlite3.connect(database=self.person_codecs.source_filename) as db:
                speaker_notes: pd.DataFrame = read_sql_table("speaker_notes", db)
                speaker_notes.set_index(self.service.id_name, inplace=True)
                return speaker_notes["speaker_note"].to_dict()
        except Exception as ex:
            logger.error(f"unable to read speaker_notes: {ex}")
            return {}

    def speech(
        self, speech_name: str, mode: Literal["dict", "text", "html"]
    ) -> dict | str:
        try:
            """Load speech data from speech corpus"""
            protocol_name: str = speech_name.split("_")[0]
            speech_nr: int = int(speech_name.split("_")[1])

            metadata, utterances = self.source.load(protocol_name)
            speech: dict = self.service.nth(
                metadata=metadata, utterances=utterances, n=speech_nr - 1
            )

            speech_info: dict = self._get_speech_info(speech_name)
            speech.update(**speech_info)
            speech.update(protocol_name=protocol_name)

            speech["office_type"] = self.person_codecs.office_type2name.get(
                speech["office_type_id"], "unknown"
            )
            speech["sub_office_type"] = self.person_codecs.sub_office_type2name.get(
                speech["sub_office_type_id"], "unknown"
            )
            speech["gender"] = self.person_codecs.gender2name.get(
                speech["gender_id"], "unknown"
            )
            speech["party_abbrev"] = self.person_codecs.party_abbrev2name.get(
                speech["party_id"], "unknown"
            )

        except Exception as ex:  # pylint: disable=bare-except
            speech = {"name": "speech not found", "error": str(ex)}

        if mode == "html":
            return self.to_html(speech)

        if mode == "text":
            return self.to_text(speech)

        return speech

    def to_text(self, speech: dict) -> str:
        paragraphs: list[str] = speech.get("paragraphs", [])
        text: str = self.fix_whitespace("\n".join(paragraphs))
        return text

    def fix_whitespace(self, text: str) -> str:
        return self.subst_puncts.sub(r"\1", text)

    def to_html(self, speech: dict) -> str:
        try:
            speech["parlaclarin_links"] = self.to_parla_clarin_urls(
                speech["protocol_name"]
            )
            speech["wikidata_link"] = self.to_wikidata_link(speech["who"])
            speech["kb_labb_link"] = self.to_kb_labb_link(
                speech["protocol_name"], speech["page_number"]
            )
            return self.template.render(speech)
        except Exception as ex:
            return f"render failed: {ex}"

    def to_parla_clarin_urls(self, protocol_name: str, ignores: str = "alpha") -> str:
        return " ".join(
            map(
                lambda x: f'<a href="{x.url}" target="_blank" style="font-weight: bold;color: blue;">{x.name}</a>&nbsp;',
                self.get_github_xml_urls(protocol_name, ignores=ignores),
            )
        )

    def to_wikidata_link(self, who: str) -> str:
        if not bool(who) or who == "unknown":
            return ""
        height, width = 20, int(20 * 1.41)
        img_src = f'<img width={width} heigh={height} src="https://upload.wikimedia.org/wikipedia/commons/f/ff/Wikidata-logo.svg"/>'
        return f'<a href="https://www.wikidata.org/wiki/{who}" target="_blank" style="font-weight: bold;color: blue;">{img_src}</a>&nbsp;'

    def to_kb_labb_link(self, protocol_name: str, page_number: str) -> str:
        if not bool(protocol_name):
            return ""

        page_url: str = (
            f"{protocol_name.replace('-', '_')}-{str(page_number).zfill(3)}.jp2/"
            if page_number.isnumeric()
            else ""
        )

        url: str = f"https://betalab.kb.se/{protocol_name}/{page_url}_view"

        return f'<a href="{url}" target="_blank" style="font-weight: bold;color: blue;">KB</a>&nbsp;'

    def get_github_tags(self, github_access_token: str = None) -> list[str]:
        release_tags: list[str] = ["main", "dev"]
        try:
            access_token: str = github_access_token or os.environ.get(
                "GITHUB_ACCESS_TOKEN", None
            )

            # if access_token is None:
            #    logger.info("GITHUB_ACCESS_TOKEN not set")

            github: gh.Github = gh.Github(access_token)

            riksdagen_corpus = github.get_repo(
                "welfare-state-analytics/riksdagen-corpus"
            )
            release_tags = release_tags + [
                x.title for x in riksdagen_corpus.get_releases()
            ]

        except:  # pylint: disable=bare-except
            ...
        return release_tags

    def get_github_xml_urls(
        self, protocol_name: str, ignores: str = None, n: int = 2
    ) -> list[GithubUrl]:
        protocol_name: str = pu.strip_extensions(protocol_name)
        sub_folder: str = protocol_name.split("-")[1]
        xml_urls: list[GithubUrl] = []
        tags: list[str] = (
            [t for t in self.release_tags if ignores not in t]
            if ignores
            else self.release_tags
        )
        for tag in tags[:n]:
            url: str = f"{self.GITHUB_REPOSITORY_URL}/blob/{tag}/corpus/protocols/{sub_folder}/{protocol_name}.xml"
            raw_url: str = f"{self.GITHUB_REPOSITORY_RAW_URL}/{tag}/corpus/protocols/{sub_folder}/{protocol_name}.xml"
            xml_urls.append(GithubUrl(name=tag, url=url))
            xml_urls.append(GithubUrl(name=f"({tag})", url=raw_url))
        return xml_urls



from typing import Any, List

import pandas as pd
import streamlit as st

import swedeb_demo.components.component_texts as ct
from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore
from swedeb_demo.components.meta_data_display import MetaDataDisplay  # type: ignore
from swedeb_demo.components.speech_display_mixin import ExpandedSpeechDisplay
from swedeb_demo.components.table_results import TableDisplay
from swedeb_demo.components.tool_tab import ToolTab


class KWICDisplay(ExpandedSpeechDisplay, ToolTab):
    def __init__(
        self, another_api: ADummyApi, shared_meta: MetaDataDisplay, tab_key: str
    ) -> None:
        super().__init__(another_api, shared_meta, tab_key)
        self.labels = ct.kwic_labels
        self.column_names = ct.kwic_col_names
        self.add_containers()

        self.CURRENT_PAGE = f"current_page_{self.TAB_KEY}"
        self.SEARCH_PERFORMED = f"search_performed__{self.TAB_KEY}"
        self.EXPANDED_SPEECH = f"expanded_speech__{self.TAB_KEY}"
        self.DATA_KEY = f"data_{self.TAB_KEY}"
        self.SORT_KEY = f"sort_key_{self.TAB_KEY}"
        self.ASCENDING_KEY = f"ascending_{self.TAB_KEY}"
        self.SEARCH_BOX = f"search_box_{self.TAB_KEY}"
        self.N_WORDS_BEFORE = f"n_words_before_{self.TAB_KEY}"
        self.N_WORDS_AFTER = f"n_words_after_{self.TAB_KEY}"
        self.LEMMA_WORD_TOGGLE = f"lemma_word_toggle_{self.TAB_KEY}"

        if self.has_and_is(self.EXPANDED_SPEECH):
            self.display_expanded_speech(
                self.get_reset_dict(),
                self.api,
                self.TAB_KEY,
                st.session_state["selected_hit"],
            )
        else:
            self.define_displays()

            with self.top_container:
                self.draw_search_settings()

            if self.has_and_is(self.SEARCH_PERFORMED):
                self.show_display()
            else:
                self.init_session_state(self.get_initial_values())


    def draw_search_settings(self):
        with st.form(key=f"form_{self.TAB_KEY}"):
            st.caption(ct.kwic_desc)
            st.text_input(ct.kwic_text_input, key=self.SEARCH_BOX)
            self.add_window_size()
            self.add_lemma_word_toggle()
            button_name = ct.kwic_search_button
            if self.has_and_is(self.SEARCH_PERFORMED):
                button_name = ct.kwic_update_button
            st.form_submit_button(button_name, on_click=self.handle_button_click)
        self.draw_line()


    def get_st_dict_when_button_clicked(self) -> dict:
        return {
            self.SEARCH_PERFORMED: True,
            self.CURRENT_PAGE: 0,
            self.EXPANDED_SPEECH: False,
        }

    def add_containers(self):
        self.top_container = st.container()
        self.result_desc_container = st.container()
        self.n_hits_container = st.container()
        self.result_container = st.container()

    def add_lemma_word_toggle(self) -> None:
        st.toggle(ct.kwic_toggle, key=self.LEMMA_WORD_TOGGLE, help=ct.kwic_toggle_help)

    def get_initial_values(self) -> dict:
        return {
            #self.SEARCH_PERFORMED: False,
            self.CURRENT_PAGE: 0,
        }

    def handle_button_click(self) -> None:
        if self.get_search_box().strip() == "":
            with self.top_container:
                st.warning("Fyll i en sökterm")
            st.session_state[self.SEARCH_PERFORMED] = False
            st.write()  # hack to stay in kwic tab
        else:
            self.handle_search_click(self.get_st_dict_when_button_clicked())

    def add_window_size(self) -> None:
        cols_before, cols_after, _ = st.columns([2, 2, 2])
        with cols_before:
            self.add_window_select(ct.kwic_word_before, self.N_WORDS_BEFORE)
        with cols_after:
            self.add_window_select(ct.kwic_words_after, self.N_WORDS_AFTER)

    def add_window_select(self, message: str, key: str):
        st.number_input(
            message,
            key=key,
            min_value=0,
            max_value=5,
            value=st.session_state[key] if key in st.session_state else 2,
        )

    def define_displays(self) -> None:
        self.table_display = TableDisplay(
            current_container_key=self.TAB_KEY,
            current_page_name=self.CURRENT_PAGE,
            party_abbrev_to_color=self.api.party_abbrev_to_color,
            expanded_speech_key=self.EXPANDED_SPEECH,
            table_type=ct.kwic_table_type,
            data_key=self.DATA_KEY,
        )

    def get_reset_dict(self) -> dict:
        return {
            self.SEARCH_PERFORMED: True,
            self.CURRENT_PAGE: st.session_state[self.CURRENT_PAGE],
            self.SEARCH_BOX: self.get_search_box(),
            self.LEMMA_WORD_TOGGLE: self.get_lemma_word_toggle(),
            self.N_WORDS_BEFORE: self.get_n_words_before(),
            self.N_WORDS_AFTER: self.get_n_words_before(),
            self.EXPANDED_SPEECH: False,
        }

    def get_lemma_word_toggle(self):
        if self.LEMMA_WORD_TOGGLE in st.session_state:
            return st.session_state[self.LEMMA_WORD_TOGGLE]
        return False
    
    def get_n_words_after(self):
        if self.N_WORDS_AFTER in st.session_state:
            return st.session_state[self.N_WORDS_AFTER]
        return 2

    def get_n_words_before(self):
        if self.N_WORDS_BEFORE in st.session_state:
            return st.session_state[self.N_WORDS_BEFORE]
        return 2

    def show_display(self) -> None:
        hit = self.get_search_box()
        if hit:
            hits = [h.strip().lower() for h in hit.split(" ")]
            if not self.has_and_is(self.LEMMA_WORD_TOGGLE):
                hits = [h.lower() for h in hit.split(" ")]
            self.show_hit(hits)
        else:
            self.display_settings_info_no_hits()

    def show_hit(self, hits: List[str]) -> None:
        selections = self.search_display.get_selections()
        data = self.get_data(
            hits,
            self.search_display.get_slider(),
            selections=selections,
            words_before=st.session_state[self.N_WORDS_BEFORE],
            words_after=st.session_state[self.N_WORDS_AFTER],
            lemmatized=not st.session_state[self.LEMMA_WORD_TOGGLE],
        )

        if data.empty:
            self.display_settings_info_no_hits()
        else:
            st.session_state[self.DATA_KEY] = data
            with self.n_hits_container:
                self.add_download_button(data, ct.kwic_filename)
            with self.result_desc_container:
                self.display_settings_info(n_hits=len(data))
            with self.result_container:
                st_columns = self.table_display.get_kwick_columns()
                self.add_sort_buttons(self.labels, st_columns[:-1], self.column_names)

                with st_columns[-1]:
                    st.markdown(ct.kwic_show_speeches)

                if self.SORT_KEY in st.session_state:
                    st.session_state[self.DATA_KEY] =  data.sort_values(
                        st.session_state[self.SORT_KEY],
                        ascending=st.session_state[self.ASCENDING_KEY],
                       
                    )
                self.table_display.write_table()

    @st.cache_data
    def get_data(
        _self,
        hits: List[str],
        slider: Any,
        selections: dict,
        words_before: int,
        words_after: int,
        lemmatized: bool = True,
    ) -> pd.DataFrame:
        data = _self.api.get_kwic_results_for_search_hits(
            hits,
            from_year=slider[0],
            to_year=slider[1],
            selections=selections,
            words_before=words_before,
            words_after=words_after,
            lemmatized=lemmatized,
        )

        return data



from typing import Any, List, Tuple

import pandas as pd
import plotly.graph_objects as go  # type: ignore
import streamlit as st

import swedeb_demo.components.component_texts as ct
from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore
from swedeb_demo.components.meta_data_display import \
    MetaDataDisplay  # type: ignore
from swedeb_demo.components.speech_display_mixin import ExpandedSpeechDisplay
from swedeb_demo.components.table_results import TableDisplay
from swedeb_demo.components.tool_tab import ToolTab


class WordTrendsDisplay(ExpandedSpeechDisplay, ToolTab):
    def __init__(
        self, another_api: ADummyApi, shared_meta: MetaDataDisplay, tab_key: str
    ) -> None:
        super().__init__(another_api, shared_meta, tab_key)

        self.labels = ct.wt_table_labels
        self.column_names = ct.wt_column_names

        self.SEARCH_PERFORMED = f"search_performed_{self.TAB_KEY}"
        self.CURRENT_PAGE_TAB = f"current_page_tab_{self.TAB_KEY}"
        self.CURRENT_PAGE_SOURCE = f"current_page_source_{self.TAB_KEY}"
        self.EXPANDED_SPEECH = f"expanded_speech_{self.TAB_KEY}"
        self.NORMAL_TABLE_WT = f"normalize_table_{self.TAB_KEY}"
        self.NORMAL_DIAGRAM_WT = f"normalize_diagram_{self.TAB_KEY}"
        self.DISPLAY_SELECT = f"display_select_{self.TAB_KEY}"
        self.SORT_KEY = f"sort_key_{self.TAB_KEY}"
        self.ASCENDING_KEY = f"ascending_{self.TAB_KEY}"
        self.DATA_KEY_TABLE = f"data_table_{self.TAB_KEY}"
        self.DATA_KEY_SOURCE = f"data_source_{self.TAB_KEY}"
        self.SEARCH_BOX = f"search_box_{self.TAB_KEY}"
        self.words_per_year = self.api.get_words_per_year()
        self.NON_NORMALIZED = "Absolut frekvens"
        self.NORMALIZED = "Normaliserad frekvens"
        self.ANFORANDEN = ct.wt_option_anforanden
        self.TABELL = ct.wt_option_tabell
        self.DIAGRAM = ct.wt_option_diagram

        if self.has_and_is(self.EXPANDED_SPEECH):
            self.display_expanded_speech(
                self.get_reset_dict(),
                self.api,
                self.TAB_KEY,
                st.session_state['selected_hit'],
            )
        
        else:
            st.caption(ct.word_trend_desc)

            self.top_container = st.container()
            self.search_container = st.container()
            self.top_result_container = st.container()
            self.result_container = st.container()

            self.st_dict_when_button_clicked = {
                self.SEARCH_PERFORMED: True,
                self.CURRENT_PAGE_TAB: 0,
                self.CURRENT_PAGE_SOURCE: 0,
                self.EXPANDED_SPEECH: False,

            }

            self.define_displays()

            with self.search_container:
                self.draw_search_settings()
                

            if self.has_and_is(self.SEARCH_PERFORMED):
                self.show_display()
            else:
                self.init_session_state(self.get_initial_values())

    def draw_search_settings(self):
        with st.form(key=f"form_{self.TAB_KEY}"):
            st.text_input(ct.wt_text_input, key=self.SEARCH_BOX)
            st.form_submit_button(ct.wt_search_button, 
                                  on_click=self.handle_button_click)
        self.draw_line()


    def get_initial_values(self):
        session_state_initial_values = {
            #self.SEARCH_PERFORMED: False,
            self.CURRENT_PAGE_TAB: 0,
            self.CURRENT_PAGE_SOURCE: 0,
            self.NORMAL_TABLE_WT: self.NON_NORMALIZED,
            self.NORMAL_DIAGRAM_WT: self.NON_NORMALIZED,
            self.EXPANDED_SPEECH: False,
        }

        return session_state_initial_values

    def get_reset_dict(self) -> dict:
        return {
            self.SEARCH_PERFORMED: True,
            self.CURRENT_PAGE_SOURCE: st.session_state[self.CURRENT_PAGE_SOURCE],
            self.SEARCH_BOX: self.get_current_search_as_str(),
            self.DISPLAY_SELECT: self.ANFORANDEN,
            self.EXPANDED_SPEECH: False,
        }

    def get_current_search_as_str(self) -> str:
        if self.SEARCH_BOX in st.session_state:
            return st.session_state[self.SEARCH_BOX]
        return ""

    def handle_button_click(self) -> None:
        if self.get_search_box().strip() == "":
            st.warning("Fyll i en sökterm")
            st.session_state[self.SEARCH_PERFORMED] = False
        else:
            self.handle_search_click(self.st_dict_when_button_clicked)

    def define_displays(self) -> None:
        self.table_display_table = TableDisplay(
            current_container_key=self.TAB_KEY,
            current_page_name=self.CURRENT_PAGE_TAB,
            party_abbrev_to_color=self.api.party_abbrev_to_color,
            expanded_speech_key="",
            table_type=ct.wt_table_type,
            data_key=self.DATA_KEY_TABLE,
        )
        self.table_display = TableDisplay(
            current_container_key=self.TAB_KEY,
            current_page_name=self.CURRENT_PAGE_SOURCE,
            party_abbrev_to_color=self.api.party_abbrev_to_color,
            expanded_speech_key=self.EXPANDED_SPEECH,
            table_type=ct.wt_source_type,
            data_key=self.DATA_KEY_SOURCE,
        )

    @st.cache_data
    
    def get_data(
        _self,
        search_words: List[str],
        start_year: int,
        end_year: int,
        selections: dict,
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        st.session_state['word_trend_selections'] = selections
        res = _self.api.get_word_trend_results(
            search_words,
            filter_opts=selections,
            start_year=start_year,
            end_year=end_year,
        )

        if res[0].shape[1] > 1:
            res[0]['Totalt'] = res[0].sum(axis=1)
        res[0].reset_index(inplace=True)
        res[0].rename(columns={"year": "År"}, inplace=True)
        res[0].set_index(res[0].columns[0], inplace=True)
        return res
    
    def normalize_word_per_year(self, data: pd.DataFrame) -> pd.DataFrame:
        data = data.merge(self.words_per_year, left_index=True, right_index=True)
        data = data.iloc[:, :].div(data.n_raw_tokens, axis=0)
        data.drop(columns=["n_raw_tokens"], inplace=True)

        return data

    def show_display(self) -> None:
        slider = self.search_display.get_slider()

        selections = self.search_display.get_selections()
        hits = self.get_search_terms()

        if len(hits) > 0:
            with self.search_container:
                hit_selector = st.multiselect(
                    label=ct.wt_hit_selector, options=hits, default=hits
                )

            data, kwic_like_data = self.get_data(
                hit_selector,
                slider[0],
                slider[1],
                selections,
            )

            with self.top_result_container:
                self.draw_line()
                if kwic_like_data.empty:
                    self.display_settings_info_no_hits()
                else:
                    self.display_settings_info(n_hits=len(kwic_like_data), 
                                               hits=f": {', '.join(hits)}")

                    col_display_select, down_col_a, down_col_b = st.columns([2, 1, 1])
                    with col_display_select:
                        self.add_radio_buttons()
                    with down_col_a:
                        st.write("")
                        self.add_download_button(data, ct.wt_filename, button_label="Ladda ner ordfrekvenser", index=True)
                    with down_col_b:
                        st.write("")
                        self.add_download_button(kwic_like_data, ct.wt_filename_speeches, button_label="Ladda ner anföranden")
                    self.draw_line()
                    self.display_results(data, kwic_like_data)
        else:
            self.display_settings_info_no_hits(self)

    def add_radio_buttons(self):
        st.radio(
            ct.wt_options_desc,
            ct.wt_result_options,
            index=0,
            key=self.DISPLAY_SELECT,
            help=None,
            horizontal=True,
        )

    def get_search_terms(self):
        search_terms = self.get_search_box().split(",")
        search_terms = [term.strip().lower() for term in search_terms]

        hits = self.parse_search_string(search_terms)
        return hits

    def parse_search_string(self, search_terms):
        hits = []
        if len(search_terms) > 0:
            for term in search_terms:
                current_hits = self.api.get_word_hits(term, n_hits=10)
                for hit in current_hits:
                    hits.append(hit)
                hits.extend(current_hits)
        hits = list(set(hits))
        return hits

    def normalize(self, data, normalization_key):
        self.radio_normalize(normalization_key)
        normalize = st.session_state[normalization_key] == self.NORMALIZED
        if normalize:
            data = self.normalize_word_per_year(data)
        return data

    def display_results(
        self, data: pd.DataFrame, kwic_like_data: pd.DataFrame
    ) -> None:
        with self.result_container:
            if st.session_state[self.DISPLAY_SELECT] == self.DIAGRAM:
                data = self.normalize(data, self.NORMAL_DIAGRAM_WT)
                self.draw_line_figure(data)

            elif st.session_state[self.DISPLAY_SELECT] == self.TABELL:
                data = self.normalize(data, self.NORMAL_TABLE_WT)
                st.session_state[self.DATA_KEY_TABLE] = data
                self.table_display_table.write_table()

            else:
                self.add_anforande_display(kwic_like_data)

    def add_anforande_display(self, kwic_like_data):
        st_columns = self.table_display.get_columns(include_hit=True)
        self.add_sort_buttons(self.labels, st_columns[:-1], self.column_names)

        with st_columns[-1]:
            st.write(ct.wt_speech_col)

        if self.SORT_KEY in st.session_state:
            kwic_like_data.sort_values(
                st.session_state[self.SORT_KEY],
                ascending=st.session_state[self.ASCENDING_KEY],
                inplace=True,
            )

        st.session_state[self.DATA_KEY_SOURCE] = kwic_like_data
        self.table_display.write_table()

    def radio_normalize(self, key):
        st.radio(
            ct.wt_norm_radio_title,
            (self.NON_NORMALIZED, self.NORMALIZED),
            key=key,
            help=ct.wt_norm_help,
            horizontal=True,
        )

    def draw_line_figure(self, data: pd.DataFrame) -> None:
        markers = ct.wt_plot_markers
        lines = ct.wt_plot_lines
        fig = go.Figure()
        for i, col in enumerate(data.columns):
            fig.add_trace(
                go.Scatter(
                    x=data.index,
                    y=data[col],
                    mode="lines+markers",
                    name=col,
                    marker=dict(symbol=markers[i % len(markers)], size=8),
                    line=dict(dash=lines[i % len(lines)]),
                    showlegend=True,
                )
            )
        fig.update_yaxes(exponentformat="none")
        fig.update_layout(xaxis_title=ct.wt_x_axis, yaxis_title=ct.wt_y_asix)
        fig.update_layout(separators=",   ")
        st.plotly_chart(fig, use_container_width=True, theme="streamlit")



import io
import zipfile
from typing import List

import pandas as pd


def convert(source_filename: str, target_filename: str, compresslevel: int = 9):
    with zipfile.ZipFile(source_filename, "r") as sfp:
        with zipfile.ZipFile(
            target_filename,
            "w",
            compression=zipfile.ZIP_DEFLATED,
            compresslevel=compresslevel,
        ) as tfp:
            filenames: List[str] = sfp.namelist()

            for filename in filenames:
                tagged_frame: pd.DataFrame = pd.read_csv(
                    io.StringIO(sfp.read(filename).decode("utf-8")),
                    sep="\t",
                    index_col=None,
                )

                tagged_frame["baseform"] = tagged_frame.apply(
                    lambda x: x.get("hist.baseform")
                    if x.get("hist.baseform")
                    else x.get("saldo.baseform")
                )

                data: str = tagged_frame[["token", "pos", "baseform"]].to_csv(
                    sep="\t", index=False
                )

                tfp.writestr(filename, data)



######################
# KWIC display texts #
######################
kwic_desc = """Med verktyget **Key Words in Context** kan du söka på ord och fraser, 
t ex `information` eller `information om`, och se kontexten till vänster och 
höger om sökningen. För att få fler träffar kan `.*` användas, 
t ex `information.*`. Under Filtrera sökresultat kan du avgränsa 
anförandena till vissa partier, talare eller år. Observera att denna test-korpus är 
lemmatiserad, dvs sökresultateten baseras på ordets grammatiska rot."""
kwic_toggle = "Exakt match (ej lemmatiserat)"
kwic_word_before = "Antal ord före sökordet (0–5 ord)"
kwic_words_after = "Antal ord efter sökordet (0–5 ord)"
kwic_toggle_help = (
    "Lemmatisering ger fler träffar eftersom "
    "även böjningsformer av sökordet inkluderas"
)
# KWIC table
kwic_table_type = "kwic"

# KWIC download filename
kwic_filename = "kwic.csv"

# KWIC table display
kwic_labels = ["Vänster", "Träff", "Höger", "Parti", "År", "Talare", "Kön"]
kwic_col_names = [
    "Kontext Vänster",
    "Sökord",
    "Kontext Höger",
    "Parti",
    "År",
    "Talare",
    "Kön",
]

# KWIC search button
kwic_search_button = "Sök"
kwic_update_button = "Uppdatera sökning"
kwic_text_input = "Skriv sökterm:"

# KWIC show speeches
kwic_show_speeches = "Visa anföranden"

###############################
# WORD TRENDs display texts   #
###############################

word_trend_desc = ("Sök på ett eller flera ord för att se hur de har använts över tid. "
"För att söka på flera ord särskilj med kommatecken, t ex `debatt,information`. "
"Sök med `*` för att få fler varianter, t.ex. `debatt*`. Under Filtrera sökresultat " 
"kan du avgränsa anförandena till vissa partier, talare eller år. Observera att denna" 
"test-korpus är lemmatiserad, dvs sökresultateten baseras på ordets grammatiska rot.")

# word trends table/plot/speeches
wt_table_type = "table"
wt_source_type = "source"
wt_speech_col = "Tal"

# word trends search settings
wt_hit_selector = "Välj sökord att inkludera"
wt_option_tabell = "Tabell"
wt_option_diagram = "Diagram"
wt_option_anforanden = "Anföranden"
wt_result_options = [wt_option_diagram, wt_option_tabell, wt_option_anforanden]
wt_options_desc = "Visa resultat som:"
wt_text_input = "Skriv sökterm:"
wt_search_button = "Sök"

# word trends normalization
wt_norm_radio_title = "Normalisera resultatet?"
wt_norm_help = """Frekvens: antal förekomster av söktermen per år. Normaliserad 
frekvens: antal förekomster av söktermen delat med totalt antal ord i tal 
under samma år."""

# word trends table display
wt_table_labels = ["Talare↕", "Kön↕", "År↕", "Parti↕", "Källa↕", "Träff↕"]
wt_column_names = ["Talare", "Kön", "År", "Parti", "Protokoll", "hit"]

# word trends plot settings
wt_plot_markers = ["circle", "hourglass", "x", "cross", "square", 5]
wt_plot_lines = ["solid", "dash", "dot", "dashdot", "solid"]
wt_x_axis = "År"
wt_y_asix = "Frekvens"

# word trends download filename
wt_filename = "Ordtrender_frekvens.csv"
wt_filename_speeches = "Ordtrender_anforanden.csv"


###############################
# SPEECHES display texts      #
###############################

sp_desc = (
    "Sök på hela anföranden. Under Filtrera sökresultat kan du avgränsa anförandena "
    "till vissa partier, talare eller år. Observera att du i dagsläget endast "
    "kan ladda ner en lista med metadata om anföranden och inte talen "
    "i sig (men det kommer man kunna göra i den färdiga versionen)."
)

# speeches table
sp_table_type = "source"

# speeches table display
sp_labels = ["Talare↕", "År↕", "Kön↕", "Parti↕", "Källa↕"]
sp_col_names = ["Talare", "År", "Kön", "Parti", "Protokoll"]

# speeches search settings
sp_search_button = "Visa anföranden"

# speeches download filename
sp_filename = "anforanden.csv"

###############################
# Main page display texts     #
###############################

# header
m_title = "Svenska riksdagsdebatter"

# corpus selectbox
m_corpus_selectbox = "Välj korpus"
m_corpus_selectbox_help = "Välj vilket korpus du vill arbeta med."
m_corpus_selectbox_options = ["Riksdagsanföranden 1960–1969, lemmatiserade"]

# meta sidebar settings
m_meta_header = "Filtrera sökresultat"
m_meta_help = """Filtrera sökresultatet efter metadata, t.ex. kön, parti, och 
tidsperiod. Filtreringen påverkar alla verktyg."""
m_meta_caption = "Filtreringen påverkar resultaten för alla verktyg"
m_meta_expander = "Filtreringsalternativ"

# hits per page
m_hits_per_page = "Antal resultat per sida"
m_hits_options = [5, 10, 20, 50]

# tabs
m_kwic_tab = "KWIC"
m_wt_tab = "WT"
m_sp_tab = "SPEECH"
m_ngram_caption = """Under utveckling. I den färdiga versionen kommer det gå att 
utforska olika fraser, dess frekvenser och kontexter."""
m_topics_caption = """Under utveckling. I den färdiga versionen kommer det gå att 
utforska olika temamodeller (eng ”topic models”) och t ex följa teman över tid och 
dess relation till varandra."""

# about tab

humlab_link = "https://www.umu.se/humlab/"
humlab = f"[Humlab]({humlab_link})"
swerik_link = "https://swerik-project.github.io"
swerik = f"[Swerik]({swerik_link})"
fredrik_link = "https://mau.se/personer/fredrik.noren/"
fredrik = f"[Fredrik Mohammadi Norén]({fredrik_link})"

m_about_caption = (
    "SweDeb (Swedish Parliamentary Debates) är ett "
    "infrastrukturprojekt "
    f"som finansieras av Umeå universitet (2023–2024) och utvecklas av {humlab} – "
    "universitetets center för digital humaniora. Syftet med SweDeb är att ta "
    "fram ett användarvänligt gränssnitt som forskare och studenter kan "
    "använda för att utforska alla riksdagsanföranden från 1867 och framåt. "
    "Det färdiga gränssnittet ska lanseras i slutet av 2024.\n\n"
    "Det underliggande datasetet till SweDeb (dvs riksdagsanförandena) hämtas "
    f"från det RJ-finansierade infrastrukturprojektet {swerik}, vars syfte bl a "
    "är att (1) skapa "
    "en databas med alla riksdagsledamöter sedan 1867, (2) annotera kammarens "
    "protokoll "
    "och märka upp alla enskilda anföranden för att sedan (3) koppla samman "
    "varje anförande med aktuell ledamot.\n\n"
    f"För frågor om SweDeb eller Swerik kontakta projektledare {fredrik} "
)


###############################
# Gender checkboxes texts     #
###############################

g_hint = "Välj kön i menyn till vänster för att visa resultat för enskilda grupper  \n"



from __future__ import annotations

import os
from typing import List, Mapping,  Union

import pandas as pd
import penelope.utility as pu  # type: ignore
from ccc import Corpora, Corpus
from dotenv import load_dotenv
from penelope.common.keyness import KeynessMetric  # type: ignore
from penelope.corpus import VectorizedCorpus  # type: ignore
from penelope.utility import PropertyValueMaskingOpts  # type: ignore

from swedeb_demo.api.parlaclarin.trends_data import (SweDebComputeOpts,
                                                     SweDebTrendsData)
from swedeb_demo.api.westac.riksprot.parlaclarin import codecs as md
from swedeb_demo.api.westac.riksprot.parlaclarin import speech_text as sr


class ADummyApi:
    """Dummy API for testing and developing the SweDeb GUI"""

    def __init__(self, env_file: str = ".env_sample_docker", 
                 corpus_dir = "/usr/local/share/cwb/registry/", 
                 corpus_name="RIKSPROT_V090_TEST") -> None:
        load_dotenv(env_file)
        self.tag: str = os.getenv("TAG")
        self.folder = os.getenv("FOLDER")
        METADATA_FILENAME = os.getenv("METADATA_FILENAME")
        TAGGED_CORPUS_FOLDER = os.getenv("TAGGED_CORPUS_FOLDER")
        self.corpus_dir = corpus_dir
        self.corpus_name = corpus_name

        self.load_corpus()
        self.data: md.Codecs = md.Codecs().load(source=METADATA_FILENAME)

        self.person_codecs: md.PersonCodecs = md.PersonCodecs().load(
            source=METADATA_FILENAME
        )

        self.repository: sr.SpeechTextRepository = sr.SpeechTextRepository(
            source=TAGGED_CORPUS_FOLDER,
            person_codecs=self.person_codecs,
            document_index=self.corpus.document_index,
        )

        self.gender_to_swedish = {"man": "Man", "woman": "Kvinna", "unknown": "Okänt"}

        self.party_specs = self.get_party_specs()
        self.decoded_persons = self.data.decode(
            self.person_codecs.persons_of_interest, drop=False
        )
        self.possible_pivots = [
            v["text_name"] for v in self.person_codecs.property_values_specs
        ]
        self.party_id_to_color = dict(
            zip(self.data.party.index, self.data.party.party_color)
        )
        self.party_abbrev_to_color = dict(
            zip(self.data.party.party_abbrev, self.data.party.party_color)
        )
        self.kwic_corpus = self.load_kwic_corpus()
        self.words_per_year = self._set_words_per_year()
    
    def get_only_parties_with_data(self):
        parties_in_data = self.corpus.document_index.party_id.unique()
        return parties_in_data

    def _set_words_per_year(self) -> pd.DataFrame:
        data_year_series = self.corpus.document_index.groupby("year")[
            "n_raw_tokens"
        ].sum()
        return data_year_series.to_frame().set_index(data_year_series.index.astype(str))

    def get_words_per_year(self):
        return self.words_per_year

    def load_kwic_corpus(self) -> Corpus:
        corpora: Corpora = Corpora(registry_dir=self.corpus_dir)
        corpus: Corpus = corpora.corpus(corpus_name=self.corpus_name)
        return corpus

    def load_corpus(self) -> None:
        self.corpus = VectorizedCorpus.load(folder=self.folder, tag=self.tag)

    def get_party_specs(self) -> Union[str, Mapping[str, int]]:
        for specification in self.data.property_values_specs:
            if specification["text_name"] == "party_abbrev":
                specs = specification["values"]
                selected = {}
                for k, v in specs.items():
                    if v in self.get_only_parties_with_data():
                        selected[k] = v 
                return selected

    def get_word_hits(self, search_term: str, n_hits: int = 5) -> list[str]:
        search_term = search_term.lower()
        return self.corpus.find_matching_words({f"{search_term}"}, n_hits)

    def get_speech(self, document_name: str):  # type: ignore
        return self.repository.speech(speech_name=document_name, mode="dict")

    def get_speech_text(self, document_name: str):  # type: ignore
        return self.repository.to_text(self.get_speech(document_name))

    def get_speaker_note(self, document_name: str) -> str:
        speech = self.get_speech(document_name)
        if "speaker_note_id" not in speech:
            return ''
        if speech["speaker_note_id"] == "missing":
            return "Talet saknar notering"
        else:
            return speech["speaker_note"]

    def get_word_vectors(
        self, words: list[str], corpus: VectorizedCorpus = None
    ) -> dict:
        """Returns individual corpus columns vectors for each search term

        Args:
            words: list of strings (search terms)
            corpus (VectorizedCorpus, optional): current corpus in None. 
            Defaults to None.

        Returns:
            dict: key: search term, value: corpus column vector
        """
        vectors = {}
        if corpus is None:
            corpus = self.corpus

        for word in words:
            vectors[word] = corpus.get_word_vector(word)
        return vectors

    def filter_corpus(
        self, filter_dict: dict, corpus: VectorizedCorpus
    ) -> VectorizedCorpus:
        if filter_dict is not None:
            for key in filter_dict:
                corpus = corpus.filter(lambda row: row[key] in filter_dict[key])
        return corpus

    def get_anforanden(
        self,
        from_year: int,
        to_year: int,
        selections: dict,
        di_selected: pd.DataFrame = None,
    ) -> pd.DataFrame:
        """For getting a list of - and info about - the full 'Anföranden' (speeches)

        Args:
            from_year int: start year
            to_year int: end year
            selections dict: selected filters, i.e. genders, parties, and, speakers

        Returns:
            DatFrame: DataFrame with speeches for selected years and filter.
        """
        if di_selected is None:
            filtered_corpus = self.filter_corpus(selections, self.corpus)
            di_selected = filtered_corpus.document_index
        di_selected = di_selected[di_selected["year"].between(from_year, to_year)]

        return self.prepare_anforande_display(di_selected)

    def get_link(self, person_id, name):
        if name == "":
            "Okänd"
        return f"[{name}](https://www.wikidata.org/wiki/{person_id})"

    def prepare_anforande_display(
        self, anforanden_doc_index: pd.DataFrame
    ) -> pd.DataFrame:
        anforanden_doc_index = anforanden_doc_index[
            ["who", "year", "document_name", "gender_id", "party_id"]
        ]
        adi = anforanden_doc_index.rename(columns={"who": "person_id"})
        self.person_codecs.decode(adi, drop=False)
        adi["link"] = adi.apply(
            lambda x: self.get_link(x["person_id"], x["name"]), axis=1
        )
        adi.drop(columns=["person_id", "gender_id", "party_id"], inplace=True)

        # to sort unknowns to the end of the results
        sorted_adi = adi.sort_values(by="name", key=lambda x: x == "")
        return sorted_adi.rename(
            columns={
                "name": "Talare",
                "document_name": "Protokoll",
                "gender": "Kön",
                "party_abbrev": "Parti",
                "year": "År",
            }
        )

    def construct_multiword_query(search_terms):
        # [lemma="information"] [lemma="om"]
        query = ""
        for term in search_terms:
            query += f' [lemma="{term}"]'
        return query[1:]

    def get_query_from_selections(self, selections, prefix):
        query = ""
        for key, value in selections.items():
            query += f'&({prefix}.{key}="{value[0]}"'
            for v in value[1:]:
                query += f'|{prefix}.{key}="{v}"'
            query += ")"

        return query[1:]

    def get_query(self, search_terms, selection, lemmatized, prefix):
        term_query = self.get_search_query_list(search_terms, lemmatized)
        if selection:
            q = self.get_query_from_selections(selection, prefix=prefix)
            query = f"{prefix}:{term_query}::{q}"
            return query
        return term_query

    def get_search_query_list(self, search_terms, lemmatized):
        # [lemma="information"] [lemma="om"]
        search_setting = "lemma" if lemmatized else "word"
        query = ""
        for term in search_terms:
            query += f' [{search_setting}="{term}"]'
        return query[1:]

    def rename_selection_keys(self, selections):
        renames = {"gender_id": "speech_gender_id", "party_id": "speech_party_id",
                    "who":"speech_who"}
        for key, value in renames.items():
            if key in selections:
                selections[value] = selections.pop(key)
        return selections

    def get_kwic_results_for_search_hits(
        self,
        search_hits: List[str],
        from_year: int,
        to_year: int,
        selections: dict,
        words_before: int,
        words_after: int,
        lemmatized: bool,
    ) -> pd.DataFrame:
        selections = self.rename_selection_keys(selections)
        query_str = self.get_query(search_hits, selections, lemmatized, prefix="a")
        subcorpus = self.kwic_corpus.query(
            query_str, context_left=words_before, context_right=words_after
        )

        data: pd.DataFrame = subcorpus.concordance(
            # form='dataframe'
            form="kwic",  # 'simple', 'dataframes',...
            p_show=["word"],  # ['word', 'pos', 'lemma']
            
            s_show=[
                "speech_who",
                "speech_title",
                "year_title",
                "speech_party_id",
                "speech_gender_id",
            ],
            order="first",
            cut_off=100000,
            matches=None,
            slots=None,
            cwb_ids=False,
        )

        if len(data) == 0:
            return pd.DataFrame()

        renamed_selections = {
            "speech_gender_id": "gender_id",
            "speech_party_id": "party_id",
            "speech_who": "person_id",
        }

        data.reset_index(inplace=True)
        data.rename(columns=renamed_selections, inplace=True)
        data = data.astype({"gender_id": int, "party_id": int, "year_title": int})
        data = data[data["year_title"].between(from_year, to_year)]
        data = data.astype({"year_title": str})
        data = self.person_codecs.decode(data, drop=False)
        data["link"] = data.apply(
            lambda x: self.get_link(x["person_id"], x["name"]), axis=1
        )
        renamed_columns = {
            "left_word": "Kontext Vänster",
            "node_word": "Sökord",
            "right_word": "Kontext Höger",
            "year_title": "År",
            "name": "Talare",
            "party_abbrev": "Parti",
            "speech_title": "Protokoll",
            "gender": "Kön",
        }
        data.rename(columns=renamed_columns, inplace=True)

        return data[
            [
                "Kontext Vänster",
                "Sökord",
                "Kontext Höger",
                "Parti",
                "Talare",
                "År",
                "Kön",
                "Protokoll",
                "person_id",
                "link",
            ]
        ]

    def get_property_specs(self) -> list:
        return self.data.property_values_specs

    def get_word_trend_results(
        self,
        search_terms: List[str],
        filter_opts: dict,
        start_year: int,
        end_year: int,
    ) -> pd.DataFrame:
        search_terms = [x.lower() for x in search_terms if x in self.corpus.vocabulary]

        if not search_terms:
            return pd.DataFrame(), pd.DataFrame()

        trends_data: SweDebTrendsData = SweDebTrendsData(
            corpus=self.corpus, 
            person_codecs=self.person_codecs, 
            n_top=1000000  
        )
        pivot_keys = list(filter_opts.keys()) if filter_opts else []

        opts: SweDebComputeOpts = SweDebComputeOpts(
            fill_gaps=False,
            keyness=KeynessMetric.TF,
            normalize=False,
            pivot_keys_id_names=pivot_keys,
            filter_opts=PropertyValueMaskingOpts(**filter_opts),
            smooth=False,
            temporal_key="year",
            top_count=100000,
            unstack_tabular=False,
            words=search_terms,
        )

        trends_data.transform(opts)

        anforanden = self.get_anforanden_for_word_trends(search_terms, filter_opts)

        trends: pd.DataFrame = trends_data.extract(
            indices=trends_data.find_word_indices(opts)
        )

        trends = trends[trends["year"].between(start_year, end_year)]
        anforanden = anforanden[anforanden["År"].between(start_year, end_year)]
        trends["year"] = trends["year"].astype(str)

        trends.rename(columns={"who": "person_id"}, inplace=True)
        trends_data.person_codecs.decode(trends)

        if not pivot_keys:
            unstacked_trends = trends.set_index(opts.temporal_key)

        else:
            current_pivot_keys = [opts.temporal_key] + [
                x for x in trends.columns if x in self.possible_pivots
            ]
            unstacked_trends = pu.unstack_data(trends, current_pivot_keys)
        self.translate_dataframe(unstacked_trends)
        unstacked_trends = unstacked_trends.loc[:, (unstacked_trends != 0).any(axis=0)]
        return unstacked_trends, anforanden

    def get_anforanden_for_word_trends(self, search_terms, filter_opts):
        filtered_corpus = self.filter_corpus(filter_opts, self.corpus)
        vectors = self.get_word_vectors(search_terms, filtered_corpus)
        hits = []
        for word, vec in vectors.items():
            hit_di = filtered_corpus.document_index[vec.astype(bool)]
            anforanden = self.prepare_anforande_display(hit_di)
            anforanden["hit"] = word
            hits.append(anforanden)

        all_hits = pd.concat(hits)

        return all_hits

    def translate_gender_col_header(self, col: str) -> str:
        """Translates gender column names to Swedish

        Args:
            col str: column name, possibly a gender

        Returns:
            str: Swedish translation of column name if it represents a gender, 
            else the original column name
        """
        new_col = col
        if "man" in col and "woman" not in col:
            new_col = col.replace("man", "Män ")
        if "woman" in col:
            new_col = col.replace("woman", "Kvinnor ")
        if "unknown" in col:
            new_col = col.replace("unknown", "Okänt kön")
        return new_col

    def translate_dataframe(self, df: pd.DataFrame) -> pd.DataFrame:
        """Translates the (gender) columns of a data frame to Swedish

        Args:
            df DataFrame: data frame to translate
        """
        cols = df.columns.tolist()
        translations = {}
        for col in cols:
            translations[col] = self.translate_gender_col_header(col)
        df.rename(columns=translations, inplace=True)

    def get_years_start(self) -> int:
        """Returns the first year in the corpus"""
        return int(self.corpus.document_index["year"].min())

    def get_years_end(self) -> int:
        """Returns the last year in the corpus"""
        return int(self.corpus.document_index["year"].max())


# speech:
# (['speaker_note_id', 'who', 'u_id', 'paragraphs', 'num_tokens', 'num_words',
#  'page_number', 'page_number2', 'protocol_name', 'date', 'document_id', 'year', 
# 'document_name', 'filename', 'n_tokens', 'n_utterances', 'speech_index', 'gender_id', 
# 'party_id', 'office_type_id', 'sub_office_type_id', 'Adjective', 'Adverb', 
# 'Conjunction', 'Delimiter'



from typing import Any

import pandas as pd
import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore
from swedeb_demo.components import component_texts as ct
from swedeb_demo.components.meta_data_display import \
    MetaDataDisplay  # type: ignore
from swedeb_demo.components.speech_display_mixin import ExpandedSpeechDisplay
from swedeb_demo.components.table_results import TableDisplay
from swedeb_demo.components.tool_tab import ToolTab


class FullSpeechDisplay(ExpandedSpeechDisplay, ToolTab):
    def __init__(
        self, another_api: ADummyApi, shared_meta: MetaDataDisplay, tab_key: str
    ) -> None:
        super().__init__(another_api, shared_meta, tab_key)

        self.labels = ct.sp_labels
        self.column_names = ct.sp_col_names
        self.CURRENT_PAGE = f"current_page_{self.TAB_KEY}"
        self.SEARCH_PERFORMED = f"search_performed_{self.TAB_KEY}"
        self.EXPANDED_SPEECH = f"expanded_speech_{self.TAB_KEY}"
        self.DATA_KEY = f"data_{self.TAB_KEY}"
        self.SORT_KEY = f"sort_key_{self.TAB_KEY}"
        self.ASCENDING_KEY = f"ascending_{self.TAB_KEY}"

        if self.has_and_is(self.EXPANDED_SPEECH):
            reset_dict = {self.EXPANDED_SPEECH: False}
            self.display_expanded_speech(reset_dict, 
                                         self.api, 
                                         self.TAB_KEY, 
                                         search_hit=None)
        else:
            session_state_initial_values = {
                self.CURRENT_PAGE: 0,
                #self.SEARCH_PERFORMED: False,
                self.EXPANDED_SPEECH: False,
            }
            self.st_dict_when_button_clicked = {
                self.CURRENT_PAGE: 0,
                self.SEARCH_PERFORMED: True,
                self.EXPANDED_SPEECH: False,
            }

            st.caption(ct.sp_desc)
            self.top_container = st.container()
            self.bottom_container = st.container()

            self.table_display = TableDisplay(
                current_container_key=self.TAB_KEY,
                current_page_name=self.CURRENT_PAGE,
                party_abbrev_to_color=self.api.party_abbrev_to_color,
                expanded_speech_key=self.EXPANDED_SPEECH,
                table_type=ct.sp_table_type,
                data_key=self.DATA_KEY,
            )

            with self.top_container:
                self.add_search_button(ct.sp_search_button)
                self.draw_line()


            if self.has_and_is(self.SEARCH_PERFORMED):
                self.show_display()
            else:
                self.init_session_state(session_state_initial_values)

    def handle_button_click(self) -> None:
        for k, v in self.st_dict_when_button_clicked.items():
            st.session_state[k] = v

    @st.cache_data
    def get_anforanden(
        _self, _another_api: Any, from_year: int, to_year: int, selections: dict
    ) -> pd.DataFrame:
        data = _another_api.get_anforanden(from_year, to_year, selections)
        return data

    def show_display(self) -> None:
        start_year, end_year = self.search_display.get_slider()

        anforanden = self.get_anforanden(
            self.api,
            start_year,
            end_year,
            selections=self.search_display.get_selections(),
        )
        if len(anforanden) == 0:
            self.display_settings_info_no_hits(with_search_hits=False)
        else:
            self.display_results(anforanden)

    def display_results(self, anforanden: pd.DataFrame) -> None:
        with self.bottom_container:
            self.display_settings_info(n_hits=len(anforanden), with_search_hits=False)
            _, col_right = st.columns([4, 2])

            with col_right:
                self.add_download_button(anforanden, ct.sp_filename)
            self.draw_line()

            columns = self.table_display.get_columns()

            self.add_sort_buttons(self.labels, columns, self.column_names)

            if self.SORT_KEY in st.session_state:
                anforanden.sort_values(
                    st.session_state[self.SORT_KEY],
                    ascending=st.session_state[self.ASCENDING_KEY],
                    inplace=True,
                )

            st.session_state[self.DATA_KEY] = anforanden
            self.table_display.write_table()



import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore


class ExpandedSpeechDisplay:
    def display_expanded_speech(
        self,
        reset_dict: dict,
        api: ADummyApi,
        tab_key: str,
        search_hit: str = None,
    ) -> None:
        if 'selected_protocol' in st.session_state and \
            st.session_state['selected_protocol'] is not None:
            selected_protocol = st.session_state['selected_protocol']
            
            ch = "kammaren"
            chamber = f"Andra {ch}" if "ak" in selected_protocol else f"Första {ch}"

            simplified_protocol = (
                selected_protocol.split("-")[1]
                + ":"
                + selected_protocol.split("-")[5].split("_")[0]
            )
            dummy_pdf = "https://www.riksdagen.se/sv/sok/?avd=dokument&doktyp=prot"
            link = f"[{simplified_protocol}]({dummy_pdf})"
            col_1, col_3 = st.columns([3, 1])
            speaker_intro = api.get_speaker_note(selected_protocol)

            with col_1:
                info_text = f"""
                **Talare**: {st.session_state['selected_speaker']}  
                
                **År**: {st.session_state['selected_year']}  

                **Hela protokollet** {link}, {chamber}
                
                **Talarintroduktion**: {speaker_intro}

                """
                
                st.info(info_text)

            with col_3:
                st.button(
                    "Stäng och återgå",
                    key=f"close_button_{tab_key}",
                    on_click=self.reset_speech_state,
                    args=(reset_dict,),
                )
            st.write("**Hela anförandet:**")

    
            text = api.get_speech_text(st.session_state["selected_protocol"])
            # causes InvalidCharacterError: The string contains invalid characters.
            text = text.replace('<',' ').replace('>',' ')

            text = text.replace("\n", "<br><br>")
            if search_hit is not None:
                if search_hit not in text:
                    search_hit = search_hit.lower()
                    if search_hit not in text:
                        search_hit = search_hit.capitalize()
                text = text.replace(
                    search_hit,
                    f'<span style="background-color: #FFFF00">{search_hit}</span>',
                )
            st.markdown(
                f'<p style="border-width:2px; border-style:solid; border-color:#000000; padding: 1em;">{text}</p>',  # noqa: E501
                unsafe_allow_html=True,
            )
            
    def reset_speech_state(self, reset_dict: dict) -> None:
        st.session_state["selected_protocol"] = None
        for k, v in reset_dict.items():
            st.session_state[k] = v
        


"""_summary_: This is the main page of the SweDeb app with a global meta data filter and
 a tabbed interface for the different tools.
"""
from typing import Any

import click
import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore
from swedeb_demo.components import component_texts as ct
from swedeb_demo.components.kwic_tab import KWICDisplay  # type: ignore
from swedeb_demo.components.meta_data_display import \
    MetaDataDisplay  # type: ignore
from swedeb_demo.components.whole_speeches_tab import \
    FullSpeechDisplay  # type: ignore
from swedeb_demo.components.word_trends_tab import \
    WordTrendsDisplay  # type: ignore

st.set_page_config(
    page_title="SweDeb DEMO b",
    layout="wide",
)


def add_banner() -> None:
    """_summary_: Adds a banner to the top of the page containing the title and a 
    dropdown menu for selecting a corpus"""
    banner = st.container()
    with banner:
        header_col, corpus_col = st.columns([3, 2])
        with corpus_col:
            st.selectbox(
                ct.m_corpus_selectbox,
                (ct.m_corpus_selectbox_options),
                key="corpus_box",
                help=ct.m_corpus_selectbox_help,
                index=0,
            )
        with header_col:
            st.title(ct.m_title)


def set_swedish_for_selections() -> None:
    SELECT_TEXT = "Välj ett eller flera alternativ"
    multi_css = f"""
    <style>
    .stMultiSelect div div div div div:nth-of-type(2) {{visibility: hidden;}}
    .stMultiSelect div div div div div:nth-of-type(2)::before {{visibility: visible; content:"{SELECT_TEXT}";}}
    </style>
    """  # noqa: E501
    st.markdown(multi_css, unsafe_allow_html=True)


def add_meta_sidebar(api: ADummyApi) -> Any:
    sidebar_container = st.sidebar.container()

    with sidebar_container:
        st.header(ct.m_meta_header, help=ct.m_meta_help)
        side_expander = st.expander(ct.m_meta_expander, expanded=True)
        st.caption(ct.m_meta_caption)

        with side_expander:
            meta_search = MetaDataDisplay(st_dict={}, api=api)
        st.selectbox(
            ct.m_hits_per_page, ct.m_hits_options, key="hits_per_page_all", index=0
        )

    return meta_search


def add_tabs(meta_search: Any, api: ADummyApi, debug: bool) -> None:
    debug_tab = " "
    if debug:
        debug_tab = "Debug"

    (
        tab_WT,
        tab_KWIC,
        tab_whole_speeches,
        tab_NG,
        tab_topics,
        tab_about,
        tab_debug,
    ) = st.tabs(
        [
            "Ordtrender",
            "KWIC",
            "Anföranden",
            "N-gram",
            "Temamodeller",
            "Om SweDeb",
            debug_tab,
        ]
    )

    def get_debug_speech(protocoll, tab) -> None:
        with tab:
            st.write(api.get_speech(protocoll))
    

    with tab_WT:
        WordTrendsDisplay(api, shared_meta=meta_search, tab_key=ct.m_wt_tab)

    with tab_KWIC:
        KWICDisplay(api, shared_meta=meta_search, tab_key=ct.m_kwic_tab)

    with tab_NG:
        st.caption(ct.m_ngram_caption)

    with tab_whole_speeches:
        FullSpeechDisplay(api, shared_meta=meta_search, tab_key=ct.m_sp_tab)

    with tab_topics:
        st.caption(ct.m_topics_caption)

    with tab_about:
        st.markdown(ct.m_about_caption, unsafe_allow_html=True)

    if debug:
        with tab_debug:
            st.caption("Session state:")
            st.write(st.session_state)
            st.text_input("Protokollsök", key="speech_finder")
            st.button("visa protokoll", key="speech_finder_button", 
                      on_click=get_debug_speech, 
                      args=(st.session_state["speech_finder"],tab_debug))

def do_render(env_file: str, debug: bool, corpus_dir, corpus_name) -> None:
    dummy_api = ADummyApi(env_file, corpus_dir, corpus_name)
    add_banner()
    set_swedish_for_selections()
    meta_search = add_meta_sidebar(dummy_api)
    add_tabs(meta_search, dummy_api, debug)

@click.command()
@click.option(
    "--env_file",
    default=".env_sample_docker",
    help="Path to .env file with environment variables",
)
@click.option("--debug", default=True, help="Show session state info in debug tab")
@click.option(
    "--corpus_dir", default="/usr/local/share/cwb/registry/", 
    help="Path to corpora directory (registry)")
@click.option(
    "--corpus_name", default="RIKSPROT_V090_TEST", help=
    "Corpus name (eg. RIKSPROT_V090_TEST)"
)

def render_main_page(env_file: str, debug: bool, corpus_dir, corpus_name) -> None:
    do_render(env_file, debug, corpus_dir, corpus_name)


if __name__ == "__main__":
    render_main_page()  # type: ignore



import math
from typing import Any

import pandas as pd
import streamlit as st


class TableDisplay:
    def __init__(
        self,
        current_container_key: str,
        current_page_name: str,
        party_abbrev_to_color: dict,
        expanded_speech_key: str,
        table_type: str,
        data_key: str,
    ) -> None:
        self.top_container = st.container()
        self.table_container = st.container()
        self.prev_next_container = st.container()

        self.current_container = current_container_key
        self.current_page_name = current_page_name
        self.table_type = table_type
        self.data_key = data_key
        self.hits_per_page = st.session_state["hits_per_page_all"]
        self.type = type
        dummy_pdf = "https://www.riksdagen.se/sv/sok/?avd=dokument&doktyp=prot"
        self.link = f"[protokoll]({dummy_pdf})"
        self.party_colors = party_abbrev_to_color
        self.expanded_speech_key = expanded_speech_key

    def write_table(self) -> None:
        if self.data_key in st.session_state:
            current_page, max_pages = self.get_current_page(
                len(st.session_state[self.data_key])
            )
            current_df = self.get_current_df(current_page)
            with self.table_container:
                if self.table_type == "table":
                    self.display_partial_table(current_df)
                elif self.table_type == "kwic":
                    # self.display_partial_kwic(st.session_state[self.data_key])
                    self.display_partial_kwic(current_df)
                else:
                    self.display_partial_source(current_df)

            self.add_buttons(current_page, max_pages)

    def get_current_df(self, current_page):
        return st.session_state[self.data_key].iloc[
            current_page
            * self.hits_per_page : ((current_page + 1) * self.hits_per_page)
        ]

    def get_current_page(self, n_rows):
        current_page = st.session_state[self.current_page_name]
        max_pages = math.ceil(n_rows / self.hits_per_page) - 1
        if current_page > max_pages:
            st.session_state[self.current_page_name] = 0
            current_page = 0
        return current_page, max_pages

    def add_buttons(self, current_page: int, max_pages: int) -> None:
        with self.prev_next_container:
            button_col_v, _, info_col, _, button_col_h = st.columns([1, 1, 1, 1, 1])

            if current_page > 0:
                button_col_v.button(
                    "Föregående",
                    key=f"{self.current_container}_F",
                    on_click=self.decrease_page,
                )
            if current_page < max_pages:
                button_col_h.button(
                    "Nästa",
                    key=f"{self.current_container}_B",
                    on_click=self.increase_page,
                )
            info_col.caption(f"Sida {current_page + 1} av {max_pages + 1}." 
                             f" Totalt {len(st.session_state[self.data_key])} träffar.")

    def display_partial_table(self, current_df: pd.DataFrame) -> None:
        st.dataframe(current_df.style.format(thousands=" "))

    def display_partial_source(self, current_df: pd.DataFrame) -> None:
        for i, row in current_df.iterrows():
            self.write_wt_row(i, row)

    def display_partial_kwic(self, current_df: pd.DataFrame) -> None:
        for i, row in current_df.iterrows():
            self.write_kwic_row(i, row)

    def get_party_with_color(self, party: str) -> str:
        if party == "?":
            return "Metadata saknas"
        if party in self.party_colors:
            color = self.party_colors[party]
            return f'<p style="color:{color}";>{party}</p>'
        return party

    def update_speech_state(self, 
                            protocol: str, 
                            speaker: str, 
                            year: str, 
                            hit=None) -> None:
        st.session_state[self.expanded_speech_key] = True
        st.session_state["selected_protocol"] = protocol
        st.session_state["selected_speaker"] = speaker
        st.session_state["selected_year"] = year
        if hit is not None:
            st.session_state["selected_hit"] = hit

    def get_kwick_columns(self) -> Any:
        return st.columns([3, 3, 3, 2, 2, 2, 2, 2])

    def write_kwic_row(self, i: int, row: pd.Series) -> None:
        (
            left_col,
            hit_col,
            right_col,
            party_col,
            year_col,
            speaker_col,
            gender_col,
            prot_col,
            # expander_col,
        ) = self.get_kwick_columns()
        # gender_col.write(self.translate_gender(["Kön"], short=True))
        speaker = "Metadata saknas" if row["Talare"] == "" else row["link"]
        speaker_col.write(speaker)
        party_col.markdown(
            self.get_party_with_color(row["Parti"]), unsafe_allow_html=True
        )

        year_col.write(str(row["År"]))
        with prot_col:
            st.button(
                "Visa", 
                key=f"{self.current_container}_b_{i}",
                on_click=self.update_speech_state,
                args=(row["Protokoll"], speaker, row["År"], row["Sökord"]),
            )
        left_col.write(row["Kontext Vänster"])
        hit_col.markdown(f"**{row['Sökord']}**")
        right_col.write(row["Kontext Höger"])
        gender_col.write(self.translate_gender(row["Kön"], short=True))

    def write_wt_row(self, i: int, row: pd.Series) -> None:
        if 'hit' in row:
            (
            speaker_col,
            gender_col,
            year_col,
            party_col,
            link_col,
            hit_col,
            expander_col,
        ) = self.get_columns(include_hit=True)
            hit_col.write(row["hit"])
        else:
            (
                speaker_col,
                year_col,
                gender_col,
                party_col,
                link_col,
                expander_col,
            ) = self.get_columns(include_hit=False)

            
        gender_col.write(self.translate_gender(row["Kön"]))
        speaker = "Metadata saknas" if row["Talare"] == "" else row["link"]
        speaker_col.write(speaker)
        year_col.write(str(row["År"]))
        party_col.markdown(
            self.get_party_with_color(row["Parti"]), unsafe_allow_html=True
        )
        with link_col:
            st.write(
                self.link.replace(
                    "protokoll", self.translate_protocol(row["Protokoll"])
                ),
                unsafe_allow_html=True,
            )
        with expander_col:
            hit = None
            if 'hit' in row:
                hit = row['hit']
            
            st.button(
                "Visa hela",
                key=f"{self.current_container}_b_{i}",
                on_click=self.update_speech_state,
                args=(row["Protokoll"], speaker, row["År"], hit),
            )

    def translate_gender(self, gender: str, short: bool = False) -> str:
        if gender == "man":
            if short:
                return "M"
            return "Man"
        elif gender == "woman":
            if short:
                return "K"
            return "Kvinna"
        else:
            if short:
                return "?"
            return "Okänt"

    def get_columns(self, include_hit=False) -> Any:
        if include_hit:
            return st.columns([2, 2, 1, 1, 2, 2, 2])
        else:
            return st.columns([2, 2, 1, 1, 2, 2])

    def increase_page(self) -> None:
        st.session_state[self.current_page_name] += 1

    def decrease_page(self) -> None:
        st.session_state[self.current_page_name] -= 1

    def reset_page(self) -> None:
        st.session_state[self.current_page_name] = 0

    @st.cache_data
    def convert_df(_self, df: pd.DataFrame) -> bytes:
        return df.to_csv(index=False).encode("utf-8")

    def translate_protocol(self, protocol_name: str) -> str:
        split = protocol_name.split("-")
        chamber = split[3]
        chamber = chamber.replace('ak', 'Andra kammaren')
        chamber = chamber.replace('fk', 'Första kammaren')
        year = split[1]
        #return  protocol_name.split('_')[0] + ":" + chamber
        return f"{chamber} {year}:{split[5].split('_')[0]}"



from __future__ import annotations

import abc
import sqlite3
from functools import cached_property
from typing import Any, List, Mapping

import numpy as np
import pandas as pd
from penelope import utility as pu

PARTY_COLORS = [
    (0, "S", "#E8112d"),
    (1, "M", "#52BDEC"),
    (2, "gov", "#000000"),
    (3, "C", "#009933"),
    (4, "L", "#006AB3"),
    (5, "V", "#DA291C"),
    (6, "MP", "#83CF39"),
    (7, "KD", "#000077"),
    (8, "NYD", "#007700"),
    (9, "SD", "#DDDD00"),
]

PARTY_COLOR_BY_ID = {x[0]: x[2] for x in PARTY_COLORS}
PARTY_COLOR_BY_ABBREV = {x[1]: x[2] for x in PARTY_COLORS}

NAME2IDNAME_MAPPING: Mapping[str, str] = {
    "gender": "gender_id",
    "office_type": "office_type_id",
    "sub_office_type": "sub_office_type_id",
    "person_id": "pid",
}
IDNAME2NAME_MAPPING: Mapping[str, str] = pu.revdict(NAME2IDNAME_MAPPING)


def read_sql_table(table_name: str, con: Any) -> pd.DataFrame:
    return pd.read_sql(f"select * from {table_name}", con)


class IRiksprotMetaData(abc.ABC):
    NAME2IDNAME_MAPPING: Mapping[str, str] = NAME2IDNAME_MAPPING
    IDNAME2NAME_MAPPING: Mapping[int, str] = IDNAME2NAME_MAPPING

    PARTY_COLOR_BY_ID: Mapping[int, str] = PARTY_COLOR_BY_ID
    PARTY_COLOR_BY_ABBREV: Mapping[str, str] = PARTY_COLOR_BY_ABBREV

    def __init__(self, data: dict):
        self.data = data

    @staticmethod
    def load(database_filename: str) -> IRiksprotMetaData:
        data: dict = IRiksprotMetaData.load_data(database_filename=database_filename)
        riksprot_metadata: IRiksprotMetaData = (
            IRiksprotMetaData(data).slim_types().set_indexes()
        )
        return riksprot_metadata

    @staticmethod
    def load_data(database_filename: str) -> dict:
        with sqlite3.connect(database=database_filename) as db:
            data: dict = {
                "chamber": read_sql_table("chamber", db),
                "gender": read_sql_table("gender", db),
                "office_type": read_sql_table("office_type", db),
                "sub_office_type": read_sql_table("sub_office_type", db),
                "person": read_sql_table("persons_of_interest", db),
                "terms_of_office": read_sql_table("terms_of_office", db),
                "government": read_sql_table("government", db),
            }

            return data

    def set_indexes(self) -> IRiksprotMetaData:
        self.chamber.set_index("chamber_id", drop=True, inplace=True)
        self.gender.set_index("gender_id", drop=True, inplace=True)
        self.office_type.set_index("office_type_id", drop=True, inplace=True)
        self.sub_office_type.set_index("sub_office_type_id", drop=True, inplace=True)
        self.terms_of_office.set_index("terms_of_office_id", drop=True, inplace=True)
        self.person.rename_axis("pid", inplace=True)

        return self

    def slim_types(self) -> IRiksprotMetaData:
        self.person["gender_id"].fillna(0, inplace=True)
        self.person["party_id"].fillna(0, inplace=True)

        self.terms_of_office.loc[
            ~self.terms_of_office["office_type_id"].isin(self.office_type.index),
            "office_type_id",
        ] = 0
        self.terms_of_office.loc[
            ~self.terms_of_office["sub_office_type_id"].isin(
                self.sub_office_type.index
            ),
            "sub_office_type_id",
        ] = 0

        as_slim_types(self.person, ["year_of_birth", "year_of_death"], np.int16)
        as_slim_types(self.person, ["gender_id"], np.int8)
        as_slim_types(self.gender, ["gender_id"], np.int8)
        as_slim_types(self.chamber, ["chamber_id"], np.int8)
        as_slim_types(self.office_type, ["office_type_id"], np.int8)
        as_slim_types(
            self.sub_office_type, ["office_type_id", "sub_office_type_id"], np.int8
        )
        as_slim_types(
            self.terms_of_office, ["start_year", "end_year", "district_id"], np.int16
        )
        as_slim_types(
            self.terms_of_office, ["office_type_id", "sub_office_type_id"], np.int8
        )

        return self

    @property
    def chamber(self) -> pd.DataFrame:
        return self.data["chamber"]

    @property
    def person(self) -> pd.DataFrame:
        return self.data["person"]

    @property
    def terms_of_office(self) -> pd.DataFrame:
        return self.data["terms_of_office"]

    @property
    def gender(self) -> pd.DataFrame:
        return self.data["gender"]

    @property
    def office_type(self) -> pd.DataFrame:
        return self.data["office_type"]

    @property
    def sub_office_type(self) -> pd.DataFrame:
        return self.data["sub_office_type"]

    @property
    def government(self) -> pd.DataFrame:
        return self.data["government"]

    @cached_property
    def gender2name(self) -> dict:
        return self.gender["gender"].to_dict()

    @cached_property
    def gender2id(self) -> dict:
        return pu.revdict(self.gender2name)

    @cached_property
    def office_type2name(self) -> dict:
        return self.office_type["office"].to_dict()

    @cached_property
    def office_type2id(self) -> dict:
        return pu.revdict(self.office_type2name)

    @cached_property
    def sub_office_type2name(self) -> dict:
        return self.sub_office_type["description"].to_dict()

    @cached_property
    def sub_office_type2id(self) -> dict:
        return pu.revdict(self.sub_office_type2name)

    @cached_property
    def pid2person_id(self) -> dict:
        return self.person["person_id"].to_dict()

    @cached_property
    def person_id2pid(self) -> dict:
        return pu.revdict(self.pid2person_id)

    @cached_property
    def pid2person_name(self) -> dict:
        return self.person["name"].to_dict()

    @cached_property
    def party_abbrev2id(self) -> dict:
        return {}

    @cached_property
    def property_values_specs(self) -> List[Mapping[str, str | Mapping[str, int]]]:
        return [
            dict(text_name="gender", id_name="gender_id", values=self.gender2id),
            dict(
                text_name="office_type",
                id_name="office_type_id",
                values=self.office_type2id,
            ),
            dict(
                text_name="sub_office_type",
                id_name="sub_office_type_id",
                values=self.sub_office_type2id,
            ),
            dict(
                text_name="party_abbrev",
                id_name="party_id",
                values=self.party_abbrev2id,
            ),
            dict(text_name="person_id", id_name="pid", values=self.person_id2pid),
        ]

    def overload_by_member_data(
        self,
        df: pd.DataFrame,
        *,
        encoded: bool = True,
        drop: bool = True,
        columns: List[str] = None,
    ) -> pd.DataFrame:
        if "who" not in df.columns:
            raise ValueError("cannot merge member data, `who` is missing in target")

        columns = (
            columns
            if columns is not None
            else ["person_id", "gender_id", "party_id", "role_type_id"]
            if encoded
            else ["gender", "party_abbrev", "role_type"]
        )

        target: pd.DataFrame = self.encoded_members if encoded else self.members
        target = target[columns]
        xi: pd.DataFrame = df.merge(target, left_on="who", right_index=True, how="left")

        if drop:
            xi.drop(columns="who", inplace=True)

        xi = self.as_slim_types(xi)

        return xi

    @cached_property
    def encoded_members(self) -> pd.DataFrame:
        mx: pd.DataFrame = self.members
        mx["who"] = mx.index
        mx = mx.assign(
            person_id=mx["who"].apply(self.who2id.get).astype(np.int32),
            gender_id=mx["gender"].apply(self.gender2id.get).astype(np.int8),
            party_id=mx["party_abbrev"].apply(self.party_abbrev2id.get).astype(np.int8),
            role_type_id=mx["role_type"].apply(self.role_type2id.get),
        )
        mx = mx.drop(columns=["who", "gender", "party_abbrev", "role_type"])
        return mx

    def decode_members_data(self, df: pd.DataFrame, drop: bool = True) -> pd.DataFrame:
        if "role_type_id" in df.columns:
            df["role_type"] = df["role_type_id"].apply(self.role_type2name.get)
        if "gender_id" in df.columns:
            df["gender"] = df["gender_id"].apply(self.gender2name.get)
        if "party_id" in df.columns:
            df["party_abbrev"] = df["party_id"].apply(self.party_abbrev2name.get)
        if "person_id" in df.columns:
            df["who"] = df["person_id"].apply(self.id2who.get)
        if drop:
            df.drop(
                columns=["person_id", "gender_id", "party_id", "role_type_id"],
                inplace=True,
                errors="ignore",
            )
        return df


class MemberNotFoundError(ValueError):
    ...


def as_slim_types(
    df: pd.DataFrame, columns: List[str], dtype: np.dtype
) -> pd.DataFrame:
    if df is None:
        return None
    if isinstance(columns, str):
        columns = [columns]
    for column in columns:
        if column in df.columns:
            df[column] = df[column].fillna(0).astype(dtype)
    return df



from __future__ import annotations

from typing import Iterable

import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore


class PartyDropDown:
    def __init__(self, another_api: ADummyApi) -> None:
        self.api = another_api
        self.party_to_id = self.api.party_specs
        self.id_to_party = {v: k for k, v in self.party_to_id.items()}
        self.id_to_party[self.party_to_id["?"]] = "Partimetadata saknas"
        self.draw_selector()

    def draw_selector(self) -> None:
        css = self.get_css_for_all(self.party_to_id.values(), self.party_to_id.keys())
        st.markdown(css, unsafe_allow_html=True)
        self.selector = st.multiselect(
            "Välj partier",
            self.id_to_party.keys(),
            format_func=self.format_party,
            default=[],
            key="party_dropdown",
        )

    def generate_css_for_party_id(self, party_id: int, party_abbrev: str) -> str:
        color = self.api.party_id_to_color.get(party_id, "#000000")
        return f"""
        span[data-baseweb="tag"][aria-label="{party_abbrev}, close by backspace"]{{
            background-color: {color};
        }}"""
       
   

    def get_css_for_all(
        self, party_ids: Iterable[int], party_abbrevs: Iterable[str]
    ) -> str:
        css = ""
        for party_id, party_abbrev in zip(party_ids, party_abbrevs):
            css += self.generate_css_for_party_id(party_id, party_abbrev)
        return f"\n<style>{css}</style>\n"

    def format_party(self, x: int) -> str:
        return self.id_to_party[x]

    def get_selection(self) -> dict | None:
        if len(self.selector) == 0:
            return None
        return {"party_id": self.selector}

    def get_selected_parties_string(self) -> str:
        if len(self.selector) == 0:
            return "**Valda partier**: Alla. "
        party_str = ", ".join([self.id_to_party[x] for x in self.selector])
        return f"**Valda partier**: {party_str}. "



from __future__ import annotations

import streamlit as st
from swedeb_demo.components import component_texts as ct


class GenderCheckBoxes:
    def __init__(self) -> None:
        self.G_CHECK_MEN = "gender_check_men"
        self.G_CHECK_WOMEN = "gender_check_women"
        self.G_CHECK_UNK = "gender_check_unknown"
        self.G_KEY = "use_gender_filter"
        st.checkbox(
            "Filtrera på kön?",
            value=False,
            key=self.G_KEY,
            on_change=self.set_box_values,
        )
        disabled = not st.session_state[self.G_KEY]

        st.checkbox("Män", disabled=disabled, key=self.G_CHECK_MEN)
        st.checkbox("Kvinnor", disabled=disabled, key=self.G_CHECK_WOMEN)
        st.checkbox("Kön okänt", disabled=disabled, key=self.G_CHECK_UNK)

    def set_box_values(self) -> None:
        st.session_state[self.G_CHECK_UNK] = True
        st.session_state[self.G_CHECK_WOMEN] = True
        st.session_state[self.G_CHECK_MEN] = True

    def get_selection(self) -> dict | None:
        if st.session_state[self.G_KEY]:
            selected = []
            if st.session_state[self.G_CHECK_MEN]:
                selected.append(1)
            if st.session_state[self.G_CHECK_WOMEN]:
                selected.append(2)
            if st.session_state[self.G_CHECK_UNK]:
                selected.append(0)
            if selected:
                return {"gender_id": selected}
        return None

    def get_selected_genders_string(self) -> str:
        if st.session_state[self.G_KEY]:
            selected = []

            if st.session_state[self.G_CHECK_MEN]:
                selected.append("Män")

            if st.session_state[self.G_CHECK_WOMEN]:
                selected.append("Kvinnor")

            if st.session_state[self.G_CHECK_UNK]:
                selected.append("Okänt kön")

            gender_str = ", ".join(selected)
            if not selected:
                return ct.g_hint
            return f"**Valda kön**: {gender_str}.\n"
        return "**Valda kön**: Alla"



from typing import Any

import pandas as pd
import streamlit as st

from swedeb_demo.api.dummy_api import ADummyApi  # type: ignore
from swedeb_demo.components.meta_data_display import \
    MetaDataDisplay  # type: ignore


class ToolTab:
    def __init__(
        self, another_api: ADummyApi, shared_meta: MetaDataDisplay, tab_key: str
    ):
        self.api = another_api
        self.search_display = shared_meta
        self.TAB_KEY = tab_key
        self.HITS_PER_PAGE = f"{self.TAB_KEY}_hits_per_page"

    def init_session_state(self, session_dict: dict) -> None:
        for k, v in session_dict.items():
            if k not in st.session_state:
                st.session_state[k] = v

    def get_search_box(self) -> str:
        if f"search_box_{self.TAB_KEY}" not in st.session_state:
            return ""
        return st.session_state[f"search_box_{self.TAB_KEY}"]

    def handle_search_click(self, st_dict_when_button_clicked: dict) -> bool:
      
        for k, v in st_dict_when_button_clicked.items():
            st.session_state[k] = v


    def display_settings_info(
        self, n_hits:int, with_search_hits: bool = True, hits: str = ""
    ) -> None:
        settings = self.search_display.get_current_settings()
        if with_search_hits:
            searches = f'{self.get_search_box()}'
            st.info(

                f"Resultat för sökningen `{searches}` {hits}  \n\n{settings}" 
                f"\n\n**Antal träffar:** {n_hits}"
            )
        else:
            st.info(
                f"Resultat för sökningen: \n \n{settings}"
                f"\n\n**Antal träffar:** {n_hits}"
            )

    def display_settings_info_no_hits(self, with_search_hits: bool = True) -> None:
        settings = self.search_display.get_current_settings()
        searches = f'{self.get_search_box()}'
        if with_search_hits:
            st.info(
                f"Inga resultat för sökningen `{searches}`.  "
                f"\n{settings}  \n Utöka filtreringen eller försök med "
                "ett annat sökord för att få fler träffar."
            )
        else:
            st.info(
                f"Inga resultat för sökningen:  \n{settings}.  \n"
                "Utöka filtreringen för att få fler träffar."
            )

    def add_hit_selector(self, hits: list) -> Any:
        hit_selector = st.multiselect(
            label="Välj ord att inkludera",
            options=hits,
            default=hits,
        )

        return hit_selector

    def draw_line(self) -> None:
        st.markdown(
            """<hr style="height:2px;border:none;color:#111111;background-color:#111111;" /> """,  # noqa: E501
            unsafe_allow_html=True,
        )

    @st.cache_data
    def convert_df(_self, df: pd.DataFrame, index:bool) -> bytes:
        df_down = df.copy()
        if 'Protokoll' in df_down.columns:
            # remove number from protocoll string 
            df_down['Protokoll'] = df_down['Protokoll'].apply(lambda x: x.split('_')[0])
        return df_down.to_csv(index=index).encode("utf-8")

    def add_download_button(self, 
                            data: pd.DataFrame, 
                            file_name: str, 
                            button_label:str=None,
                            index:bool=False) -> None:
        st.download_button(
            label="Ladda ner som csv" if button_label is None else button_label,
            data=self.convert_df(data, index),
            file_name=file_name,
            mime="text/csv",
        )

    def get_sort_direction(self, key) -> None:
        if key not in st.session_state:
            st.session_state[key] = True
        else:
            st.session_state[key] = not st.session_state[key]
        return st.session_state[key]

    def add_hits_per_page(self, key):
        st.selectbox(
            "Antal resultat per sida",
            options=[5, 10, 20, 50],
            key=key,
        )

    def has_and_is(self, key):
        if key not in st.session_state:
            return False
        return st.session_state[key]

    def set_sorting(self, sorting_key):
        st.session_state[self.ASCENDING_KEY] = self.get_sort_direction(
            f"{sorting_key}_{self.TAB_KEY}"
        )
        st.session_state[self.SORT_KEY] = sorting_key
        self.table_display.reset_page()

    def add_search_button(self, message: str) -> None:
        st.button(
            message,
            key=f"search_button_{self.TAB_KEY}",
            on_click=self.handle_button_click,
        )

    def add_sort_button(self, text, column_name):
        st.button(
            text,
            key=f"sb_{column_name}_{self.TAB_KEY}",
            on_click=self.set_sorting,
            args=(column_name,),
        )

    def add_sort_buttons(self, texts, st_columns, column_names):
        for text, st_col, col_name in zip(texts, st_columns, column_names):
            with st_col:
                self.add_sort_button(text, col_name)
